@@ -3,6 +3,7 @@ from __future__ import annotations
 from .deployer import (
     CreateRunnerWorkloadSubCommand,
     CreateWorkloadSubCommand,
+    DeleteWorkloadsSubCommand,
     DeleteWorkloadSubCommand,
     ExecWorkloadSubCommand,
     GetWorkloadSubCommand,
@@ -15,6 +16,7 @@ __all__ = [
     "CreateRunnerWorkloadSubCommand",
     "CreateWorkloadSubCommand",
     "DeleteWorkloadSubCommand",
+    "DeleteWorkloadsSubCommand",
     "DetectDevicesSubCommand",
     "ExecWorkloadSubCommand",
     "GetWorkloadSubCommand",
