@@ -1,16 +1,13 @@
 from __future__ import annotations
 
-import contextlib
+import asyncio
 import json
 import os
-import platform
 import sys
 import time
 from argparse import REMAINDER
 from typing import TYPE_CHECKING
 
-from gpustack_runner import list_service_runners
-
 from .. import envs
 from ..deployer import (
     Container,
@@ -21,23 +18,24 @@ from ..deployer import (
     ContainerMount,
     ContainerPort,
     ContainerResources,
+    ContainerRestartPolicyEnum,
     WorkloadPlan,
     WorkloadStatus,
     WorkloadStatusStateEnum,
+    async_logs_workload,
     create_workload,
     delete_workload,
     exec_workload,
     get_workload,
     list_workloads,
-    logs_workload,
 )
-from ..detector import detect_devices, supported_backends
+from ..detector import supported_backends
 from .__types__ import SubCommand
 
 if TYPE_CHECKING:
     from argparse import Namespace, _SubParsersAction
 
-_IGNORE_ENVS = (
+_IGNORE_ENVS_PREFIX = (
     "PATH",
     "HOME",
     "LANG",
@@ -70,6 +68,14 @@ _IGNORE_ENVS = (
     "GPUSTACK_",
 )
 
+_IGNORE_ENVS_SUFFIX = (
+    "_HOME",
+    "_PATH",
+    "_VISIBLE_DEVICES",
+    "_DISABLE_REQUIRE",
+    "_DRIVER_CAPABILITIES",
+)
+
 
 class CreateRunnerWorkloadSubCommand(SubCommand):
     """
@@ -79,9 +85,11 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
     backend: str
     device: str
     port: int
+    host_network: bool
+    check: bool
+    namespace: str
     service: str
     version: str
-    namespace: str
     name: str
     volume: str
     extra_args: list[str]
@@ -110,8 +118,21 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
         deploy_parser.add_argument(
             "--port",
             type=int,
-            help="port to expose (default: 80)",
-            default=80,
+            help="port to expose",
+        )
+
+        deploy_parser.add_argument(
+            "--host-network",
+            action="store_true",
+            help="use host network (default: False)",
+            default=False,
+        )
+
+        deploy_parser.add_argument(
+            "--check",
+            action="store_true",
+            help="enable health check, needs --port (default: False)",
+            default=False,
         )
 
         deploy_parser.add_argument(
@@ -150,9 +171,11 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
         self.backend = args.backend
         self.device = args.device
         self.port = args.port
+        self.host_network = args.host_network
+        self.check = args.check
+        self.namespace = args.namespace
         self.service = args.service
         self.version = args.version
-        self.namespace = args.namespace
         self.name = f"{args.service}-{args.version}".lower().replace(".", "-")
         self.volume = args.volume
         self.extra_args = args.extra_args
@@ -162,53 +185,14 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
             raise ValueError(msg)
 
     def run(self):
-        arch_name = platform.machine()
-        match arch_name:
-            case "amd64" | "x86_64" | "AMD64":
-                arch_name = "amd64"
-            case "arm64" | "aarch64" | "ARM64":
-                arch_name = "arm64"
-            case _:
-                msg = f"Unsupported architecture: {arch_name}"
-                raise ValueError(msg)
-
-        runners = list_service_runners(
-            backend=self.backend,
-            service=self.service,
-            service_version_prefix=self.version,
-            platform=f"linux/{arch_name}",
-        )
-        if not runners:
-            msg = (
-                f"No runners found for service '{self.service}' "
-                f"with version prefix '{self.version}' "
-                f"and backend '{self.backend}'."
-            )
-            raise ValueError(msg)
-
-        runner = None
-        if len(runners[0].versions[0].backends[0].versions) > 1:
-            rt_v = "0.0"
-            if devs := detect_devices():
-                rt_v = devs[0].runtime_version
-            for v in runners[0].versions[0].backends[0].versions:
-                if v.version > rt_v:
-                    continue
-                runner = v.variants[0].platforms[0]
-                break
-        if not runner:
-            runner = (
-                runners[0].versions[0].backends[0].versions[0].variants[0].platforms[0]
-            )
-
-        print(f"Using runner image: {runner.docker_image}")
         env = [
             ContainerEnv(
                 name=name,
                 value=value,
             )
             for name, value in os.environ.items()
-            if not name.startswith(_IGNORE_ENVS)
+            if not name.startswith(_IGNORE_ENVS_PREFIX)
+            and not name.endswith(_IGNORE_ENVS_SUFFIX)
         ]
         if self.backend:
             resources = ContainerResources(
@@ -229,34 +213,43 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
                 path=self.volume,
             ),
         ]
-        execution = None
+        execution = ContainerExecution(
+            privileged=True,
+        )
         if self.extra_args:
-            execution = ContainerExecution(
-                command=self.extra_args,
-            )
-        ports = [
-            ContainerPort(
-                internal=self.port,
-            ),
-        ]
-        checks = [
-            ContainerCheck(
-                delay=30,
-                interval=10,
-                timeout=5,
-                retries=3,
-                tcp=ContainerCheckTCP(port=self.port),
-                teardown=True,
-            ),
-        ]
+            execution.command = self.extra_args
+        ports = None
+        if self.port:
+            ports = [
+                ContainerPort(
+                    internal=self.port,
+                ),
+            ]
+        checks = None
+        if self.check and self.port:
+            checks = [
+                ContainerCheck(
+                    delay=60,
+                    interval=10,
+                    timeout=5,
+                    retries=6,
+                    tcp=ContainerCheckTCP(port=self.port),
+                    teardown=True,
+                ),
+            ]
         plan = WorkloadPlan(
             name=self.name,
             namespace=self.namespace,
-            host_network=True,
+            host_network=self.host_network,
             containers=[
                 Container(
-                    image=runner.docker_image,
-                    name="default",
+                    restart_policy=(
+                        ContainerRestartPolicyEnum.NEVER
+                        if not self.check
+                        else ContainerRestartPolicyEnum.ALWAYS
+                    ),
+                    image=f"gpustack/runner:{self.backend if self.backend else 'Host'}X.Y-{self.service}{self.version}",
+                    name=self.name,
                     envs=env,
                     resources=resources,
                     mounts=mounts,
@@ -271,21 +264,30 @@ class CreateRunnerWorkloadSubCommand(SubCommand):
 
         try:
             while True:
-                st = get_workload(name=self.name, namespace=self.namespace)
-                if st and st.state == WorkloadStatusStateEnum.RUNNING:
+                st = get_workload(
+                    name=self.name,
+                    namespace=self.namespace,
+                )
+                if st and st.state not in (
+                    WorkloadStatusStateEnum.PENDING,
+                    WorkloadStatusStateEnum.INITIALIZING,
+                ):
                     break
                 time.sleep(1)
 
             print("\033[2J\033[H", end="")
-            logs_stream = logs_workload(
-                name=self.name,
-                namespace=self.namespace,
-                tail=-1,
-                follow=True,
-            )
-            with contextlib.closing(logs_stream) as logs:
-                for line in logs:
+
+            async def stream_logs():
+                logs_result = await async_logs_workload(
+                    name=self.name,
+                    namespace=self.namespace,
+                    tail=-1,
+                    follow=True,
+                )
+                async for line in logs_result:
                     print(line.decode("utf-8").rstrip())
+
+            asyncio.run(stream_logs())
         except KeyboardInterrupt:
             print("\033[2J\033[H", end="")
 
@@ -298,6 +300,8 @@ class CreateWorkloadSubCommand(SubCommand):
     backend: str
     device: str
     port: int
+    host_network: bool
+    check: bool
     namespace: str
     name: str
     image: str
@@ -328,8 +332,21 @@ class CreateWorkloadSubCommand(SubCommand):
         deploy_parser.add_argument(
             "--port",
             type=int,
-            help="port to expose (default: 80)",
-            default=80,
+            help="port to expose",
+        )
+
+        deploy_parser.add_argument(
+            "--host-network",
+            action="store_true",
+            help="use host network (default: False)",
+            default=False,
+        )
+
+        deploy_parser.add_argument(
+            "--check",
+            action="store_true",
+            help="enable health check, needs --port (default: False)",
+            default=False,
         )
 
         deploy_parser.add_argument(
@@ -368,6 +385,8 @@ class CreateWorkloadSubCommand(SubCommand):
         self.backend = args.backend
         self.device = args.device
         self.port = args.port
+        self.host_network = args.host_network
+        self.check = args.check
         self.namespace = args.namespace
         self.name = args.name
         self.image = args.image
@@ -385,7 +404,8 @@ class CreateWorkloadSubCommand(SubCommand):
                 value=value,
             )
             for name, value in os.environ.items()
-            if not name.startswith(_IGNORE_ENVS)
+            if not name.startswith(_IGNORE_ENVS_PREFIX)
+            and not name.endswith(_IGNORE_ENVS_SUFFIX)
         ]
         if self.backend:
             resources = ContainerResources(
@@ -406,34 +426,43 @@ class CreateWorkloadSubCommand(SubCommand):
                 path=self.volume,
             ),
         ]
-        execution = None
+        execution = ContainerExecution(
+            privileged=True,
+        )
         if self.extra_args:
-            execution = ContainerExecution(
-                command=self.extra_args,
-            )
-        ports = [
-            ContainerPort(
-                internal=self.port,
-            ),
-        ]
-        checks = [
-            ContainerCheck(
-                delay=30,
-                interval=10,
-                timeout=5,
-                retries=3,
-                tcp=ContainerCheckTCP(port=self.port),
-                teardown=True,
-            ),
-        ]
+            execution.command = self.extra_args
+        ports = None
+        if self.port:
+            ports = [
+                ContainerPort(
+                    internal=self.port,
+                ),
+            ]
+        checks = None
+        if self.check and self.port:
+            checks = [
+                ContainerCheck(
+                    delay=60,
+                    interval=10,
+                    timeout=5,
+                    retries=6,
+                    tcp=ContainerCheckTCP(port=self.port),
+                    teardown=True,
+                ),
+            ]
         plan = WorkloadPlan(
             name=self.name,
             namespace=self.namespace,
-            host_network=True,
+            host_network=self.host_network,
             containers=[
                 Container(
+                    restart_policy=(
+                        ContainerRestartPolicyEnum.NEVER
+                        if not self.check
+                        else ContainerRestartPolicyEnum.ALWAYS
+                    ),
                     image=self.image,
-                    name="default",
+                    name=self.name,
                     envs=env,
                     resources=resources,
                     mounts=mounts,
@@ -448,21 +477,30 @@ class CreateWorkloadSubCommand(SubCommand):
 
         try:
             while True:
-                st = get_workload(name=self.name, namespace=self.namespace)
-                if st and st.state == WorkloadStatusStateEnum.RUNNING:
+                st = get_workload(
+                    name=self.name,
+                    namespace=self.namespace,
+                )
+                if st and st.state not in (
+                    WorkloadStatusStateEnum.PENDING,
+                    WorkloadStatusStateEnum.INITIALIZING,
+                ):
                     break
                 time.sleep(1)
 
             print("\033[2J\033[H", end="")
-            logs_stream = logs_workload(
-                name=self.name,
-                namespace=self.namespace,
-                tail=-1,
-                follow=True,
-            )
-            with contextlib.closing(logs_stream) as logs:
-                for line in logs:
+
+            async def stream_logs():
+                logs_result = await async_logs_workload(
+                    name=self.name,
+                    namespace=self.namespace,
+                    tail=-1,
+                    follow=True,
+                )
+                async for line in logs_result:
                     print(line.decode("utf-8").rstrip())
+
+            asyncio.run(stream_logs())
         except KeyboardInterrupt:
             print("\033[2J\033[H", end="")
 
@@ -505,11 +543,66 @@ class DeleteWorkloadSubCommand(SubCommand):
             raise ValueError(msg)
 
     def run(self):
-        st = delete_workload(self.name)
-        if st:
-            print(f"Deleted workload '{self.name}'.")
-        else:
-            print(f"Workload '{self.name}' not found.")
+        try:
+            st = delete_workload(
+                name=self.name,
+                namespace=self.namespace,
+            )
+            if st:
+                print(f"Deleted workload '{self.name}'.")
+            else:
+                print(f"Workload '{self.name}' not found.")
+        except KeyboardInterrupt:
+            pass
+
+
+class DeleteWorkloadsSubCommand(SubCommand):
+    """
+    Command to delete all workload deployments.
+    """
+
+    @staticmethod
+    def register(parser: _SubParsersAction):
+        delete_parser = parser.add_parser(
+            "delete-all",
+            help="delete all workload deployments",
+        )
+
+        delete_parser.add_argument(
+            "--namespace",
+            type=str,
+            help="namespace of the workload",
+        )
+
+        delete_parser.add_argument(
+            "--labels",
+            type=lambda s: dict(item.split("=") for item in s.split(",")),
+            required=False,
+            help="filter workloads by labels (key=value pairs separated by commas)",
+        )
+
+        delete_parser.set_defaults(func=DeleteWorkloadsSubCommand)
+
+    def __init__(self, args: Namespace):
+        self.namespace = args.namespace
+        self.labels = args.labels
+
+    def run(self):
+        try:
+            sts: list[WorkloadStatus] = list_workloads(
+                namespace=self.namespace,
+                labels=self.labels,
+            )
+            for st in sts:
+                delete_workload(
+                    name=st.name,
+                    namespace=st.namespace,
+                )
+                print(f"Deleted workload '{st.name}'.")
+            if not sts:
+                print("No workloads found.")
+        except KeyboardInterrupt:
+            pass
 
 
 class GetWorkloadSubCommand(SubCommand):
@@ -572,7 +665,7 @@ class GetWorkloadSubCommand(SubCommand):
     def run(self):
         try:
             while True:
-                workload = get_workload(self.name)
+                workload = get_workload(self.name, self.namespace)
                 if not workload:
                     print(f"Workload '{self.name}' not found.")
                     return
@@ -722,23 +815,23 @@ class LogsWorkloadSubCommand(SubCommand):
     def run(self):
         try:
             print("\033[2J\033[H", end="")
-            logs_result = logs_workload(
-                name=self.name,
-                namespace=self.namespace,
-                tail=self.tail,
-                follow=self.follow,
-            )
-            if self.follow:
-                with contextlib.closing(logs_result) as logs:
-                    for line in logs:
-                        if isinstance(line, str):
-                            print(line.rstrip())
-                        else:
-                            print(line.decode("utf-8").rstrip())
-            elif isinstance(logs_result, str):
-                print(logs_result.rstrip())
-            else:
-                print(logs_result.decode("utf-8").rstrip())
+
+            async def stream_logs():
+                logs_result = await async_logs_workload(
+                    name=self.name,
+                    namespace=self.namespace,
+                    tail=self.tail,
+                    follow=self.follow,
+                )
+                if self.follow:
+                    async for line in logs_result:
+                        print(line.decode("utf-8").rstrip())
+                elif isinstance(logs_result, str):
+                    print(logs_result.rstrip())
+                else:
+                    print(logs_result.decode("utf-8").rstrip())
+
+            asyncio.run(stream_logs())
         except KeyboardInterrupt:
             print("\033[2J\033[H", end="")
 
@@ -864,20 +957,15 @@ def format_workloads_table(sts: list[WorkloadStatus]) -> str:
     if not sts:
         return "No workloads found."
 
-    width = 100
-
     headers = ["Name", "State", "Created At"]
-    col_widths = [
-        len(str(getattr(st, attr.lower().replace(" ", "_"))))
-        for st in sts
-        for attr in headers
-    ]
-    col_widths = [max(w, len(h)) for w, h in zip(col_widths, headers, strict=False)]
-
-    total_width = sum(col_widths) + len(col_widths) * 3 + 1
-    if total_width > width:
-        scale = (width - len(col_widths) * 3 - 1) / sum(col_widths)
-        col_widths = [int(w * scale) for w in col_widths]
+    # Calculate max width for each column based on header and content
+    col_widths = []
+    for attr in headers:
+        attr_key = attr.lower().replace(" ", "_")
+        max_content_width = max(
+            [len(str(getattr(st, attr_key))) for st in sts] + [len(attr)],
+        )
+        col_widths.append(max_content_width)
 
     lines = []
     header_line = (
@@ -892,9 +980,9 @@ def format_workloads_table(sts: list[WorkloadStatus]) -> str:
 
     for st in sts:
         row = [
-            st.name.ljust(col_widths[0]),
-            st.state.ljust(col_widths[1]),
-            st.created_at.ljust(col_widths[2]),
+            str(st.name).ljust(col_widths[0]),
+            str(st.state).ljust(col_widths[1]),
+            str(st.created_at).ljust(col_widths[2]),
         ]
         line = "| " + " | ".join(row) + " |"
         lines.append(line)
