@@ -1,7 +1,11 @@
 from __future__ import annotations
 
+import asyncio
+import atexit
+import contextlib
 import re
 from abc import ABC, abstractmethod
+from concurrent.futures.thread import ThreadPoolExecutor
 from dataclasses import dataclass, field
 from enum import Enum
 from typing import TYPE_CHECKING
@@ -9,39 +13,76 @@ from typing import TYPE_CHECKING
 from dataclasses_json import dataclass_json
 
 from .. import envs
-from ..detector import detect_backend
+from ..detector import (
+    Devices,
+    ManufacturerEnum,
+    detect_devices,
+    manufacturer_to_backend,
+)
+from .__utils__ import (
+    correct_runner_image,
+    fnv1a_64_hex,
+    is_rfc1123_domain_name,
+    safe_json,
+    safe_yaml,
+    validate_rfc1123_subdomain_name,
+)
 
 if TYPE_CHECKING:
-    from collections.abc import Generator
+    from collections.abc import AsyncGenerator, Generator
 
-_RE_RFC1123_DNS_SUBDOMAIN_NAME = re.compile(
-    r"(?!-)[a-z0-9-]{1,63}(?<!-)(\.(?!-)[a-z0-9-]{1,63}(?<!-))*",
-)
+_RE_LABEL_VALUE = re.compile(r"^(?!-)[A-Za-z0-9_.-]{0,63}(?<!-)$")
 """
-Regex for RFC 1123 DNS subdomain names, which must:
-    - contain no more than 253 characters
-    - contain only lowercase alphanumeric characters, '-' or '.'
+Regex for label values, which must:
+    - contain no more than 63 characters
+    - contain only alphanumeric characters, '_', '-', or '.'
     - start with an alphanumeric character
     - end with an alphanumeric character
 """
 
-_RE_RFC1123_DNS_LABEL_NAME = re.compile(r"(?!-)[a-z0-9-]{1,63}(?<!-)")
-"""
-Regex for RFC 1123 DNS label names, which must:
-    - contain no more than 63 characters
-    - contain only lowercase alphanumeric characters or '-'
-    - start with an alphanumeric character
-    - end with an alphanumeric character
-"""
 
-_RE_RFC1035_DNS_LABEL_NAME = re.compile(r"(?![0-9-])[a-zA-Z0-9_.-]{1,63}(?<![.-])")
-"""
-Regex for RFC 1035 DNS label names, which must:
-    - contain no more than 63 characters
-    - contain only lowercase alphanumeric characters or '-'
-    - start with an alphabetic character
-    - end with an alphanumeric character
-"""
+def validate_label_name_segment(segment: str):
+    """
+    Validate a label name segment.
+
+    Args:
+        segment:
+            The label name segment to validate.
+
+    Raises:
+        ValueError:
+            If the label name segment is invalid.
+
+    """
+    try:
+        validate_rfc1123_subdomain_name(segment)
+    except ValueError as e:
+        msg = f"Invalid label name segment'{segment}'. "
+        raise ValueError(msg) from e
+
+
+def validate_label_value(value: str):
+    """
+    Validate a label value.
+
+    Args:
+        value:
+            The label value to validate.
+
+    Raises:
+        ValueError:
+            If the label value is invalid.
+
+    """
+    if not _RE_LABEL_VALUE.match(value):
+        msg = (
+            f"Invalid label value '{value}'. "
+            "It must contain no more than 63 characters, "
+            "contain only alphanumeric characters, '_', '-', or '.', "
+            "start with an alphanumeric character, "
+            "and end with an alphanumeric character."
+        )
+        raise ValueError(msg)
 
 
 class UnsupportedError(Exception):
@@ -276,6 +317,9 @@ class ContainerMountModeEnum(str, Enum):
     Read-write many mode.
     """
 
+    def __str__(self):
+        return self.value
+
 
 @dataclass
 class ContainerMount:
@@ -332,6 +376,9 @@ class ContainerPortProtocolEnum(str, Enum):
     SCTP protocol.
     """
 
+    def __str__(self):
+        return self.value
+
 
 @dataclass
 class ContainerPort:
@@ -503,6 +550,28 @@ class ContainerCheck:
     """
 
 
+class ContainerImagePullPolicyEnum(str, Enum):
+    """
+    Enum for container image pull policies.
+    """
+
+    ALWAYS = "Always"
+    """
+    Always pull the image.
+    """
+    IF_NOT_PRESENT = "IfNotPresent"
+    """
+    Pull the image if not present.
+    """
+    NEVER = "Never"
+    """
+    Never pull the image.
+    """
+
+    def __str__(self):
+        return self.value
+
+
 class ContainerProfileEnum(str, Enum):
     """
     Enum for container profiles.
@@ -517,6 +586,9 @@ class ContainerProfileEnum(str, Enum):
     Init profile.
     """
 
+    def __str__(self):
+        return self.value
+
 
 class ContainerRestartPolicyEnum(str, Enum):
     """
@@ -536,6 +608,9 @@ class ContainerRestartPolicyEnum(str, Enum):
     Never restart the container.
     """
 
+    def __str__(self):
+        return self.value
+
 
 @dataclass
 class Container:
@@ -547,6 +622,8 @@ class Container:
             Image of the container.
         name (str):
             Name of the container.
+        image_pull_policy (ContainerImagePullPolicyEnum):
+            Image pull policy of the container.
         profile (ContainerProfileEnum):
             Profile of the container.
         restart_policy (ContainerRestartPolicyEnum | None):
@@ -572,12 +649,21 @@ class Container:
 
     image: str
     """
-    Image of the container.
+    Image of the container,
+    if GPUSTACK_RUNTIME_DEPLOY_CORRECT_RUNNER_IMAGE is enabled,
+    a gpustack-runner formatted image will be corrected if possible,
+    see `correct_runner_image` for details.
     """
     name: str
     """
     Name of the container.
     """
+    image_pull_policy: ContainerImagePullPolicyEnum = (
+        ContainerImagePullPolicyEnum.IF_NOT_PRESENT
+    )
+    """
+    Image pull policy of the container.
+    """
     profile: ContainerProfileEnum = ContainerProfileEnum.RUN
     """
     Profile of the container.
@@ -620,6 +706,32 @@ class Container:
     Health checks of the container.
     """
 
+    ##
+    ## The below are internal use only fields.
+    ##
+
+    _name_rfc1123_guard: WorkloadName | None = field(
+        default=None,
+        init=False,
+        repr=False,
+        metadata={"dataclasses_json": {"exclude": lambda _: True}},
+    )
+    """
+    Name for the workload,
+    but has been guard by rfc1123 validation for certain deployment environments,
+    internal use only.
+    """
+
+    @property
+    def name_rfc1123_guard(self) -> WorkloadName:
+        if not self._name_rfc1123_guard:
+            self._name_rfc1123_guard = (
+                f"gpustack-{fnv1a_64_hex(self.name)}"
+                if not is_rfc1123_domain_name(self.name)
+                else self.name
+            )
+        return self._name_rfc1123_guard
+
 
 @dataclass
 class WorkloadSecuritySysctl:
@@ -684,7 +796,6 @@ WorkloadNamespace = str
 Namespace for a workload.
 """
 
-
 WorkloadName = str
 """
 Name for a workload.
@@ -758,7 +869,7 @@ class WorkloadPlan(WorkloadSecurity):
     """
     Namespace for the workload.
     """
-    name: WorkloadName = "default"
+    name: WorkloadName | None = None
     """
     Name for the workload,
     it should be unique in the deployer.
@@ -789,55 +900,108 @@ class WorkloadPlan(WorkloadSecurity):
     It must contain at least one "RUN" profile container.
     """
 
+    ##
+    ## The below are internal use only fields.
+    ##
+
+    _name_rfc1123_guard: WorkloadName | None = field(
+        default=None,
+        init=False,
+        repr=False,
+        metadata={"dataclasses_json": {"exclude": lambda _: True}},
+    )
+    """
+    Name for the workload,
+    but has been guard by rfc1123 validation for certain deployment environments,
+    internal use only.
+    """
+
+    @property
+    def name_rfc1123_guard(self) -> WorkloadName:
+        if not self._name_rfc1123_guard:
+            self._name_rfc1123_guard = (
+                f"gpustack-{fnv1a_64_hex(self.name)}"
+                if not is_rfc1123_domain_name(self.name)
+                else self.name
+            )
+        return self._name_rfc1123_guard
+
     def validate_and_default(self):
         """
-        Validate the workload plan and set defaults.
+        Validate and set defaults for the workload plan.
 
         Raises:
             ValueError:
                 If the workload plan is invalid.
 
         """
-        # Validate
-        if not _RE_RFC1123_DNS_LABEL_NAME.match(self.name):
-            msg = (
-                f'Workload name "{self.name}" is invalid, '
-                "it must match RFC 1123 DNS label format"
-            )
+        # Validate workload name.
+        if not self.name:
+            msg = "Workload name is required."
             raise ValueError(msg)
-        for ln in self.labels or {}:
-            for p in ln.split("/"):
-                if not _RE_RFC1123_DNS_SUBDOMAIN_NAME.match(p):
-                    msg = (
-                        f'Workload label name "{ln}" is invalid, '
-                        "it must match RFC 1123 DNS subdomain format"
-                    )
-                    raise ValueError(msg)
-        names = [c.name for c in self.containers or []]
-        if len(names) != len(set(names)):
-            msg = "Container names must be unique in a workload."
-            raise ValueError(msg)
-        for n in names:
-            if not _RE_RFC1035_DNS_LABEL_NAME.match(n):
-                msg = (
-                    f'Container name "{n}" is invalid, '
-                    "it must match RFC 1035 DNS label format"
-                )
-                raise ValueError(msg)
-        if not any(
-            c.profile == ContainerProfileEnum.RUN for c in self.containers or []
+
+        # Validate at least one RUN profile container.
+        if not self.containers or not any(
+            c.profile == ContainerProfileEnum.RUN for c in self.containers
         ):
             msg = 'Workload must contain at least one "RUN" profile container.'
             raise ValueError(msg)
 
-        # Default
-        self.labels = self.labels or {}
+        # Validate workload labels, including label names and values.
+        for ln, lv in self.labels.items():
+            for s in ln.split("/"):
+                validate_label_name_segment(s)
+            validate_label_value(lv)
+
+        # Validate container names and images.
+        container_names = []
+        for ci, c in enumerate(self.containers):
+            if not c.name or not c.name.strip():
+                msg = f"Container at index {ci} is missing a name."
+                raise ValueError(msg)
+            if not c.image:
+                msg = f"Container '{c.name}' is missing an image."
+                raise ValueError(msg)
+            container_names.append(c.name)
+        if len(container_names) != len(set(container_names)):
+            msg = "Container names must be unique in the same workload."
+            raise ValueError(msg)
+
         for c in self.containers:
+            # Default restart policy.
             if c.profile == ContainerProfileEnum.INIT:
                 if not c.restart_policy:
                     c.restart_policy = ContainerRestartPolicyEnum.NEVER
             elif not c.restart_policy:
                 c.restart_policy = ContainerRestartPolicyEnum.ALWAYS
+            # Correct runner image if needed.
+            if envs.GPUSTACK_RUNTIME_DEPLOY_CORRECT_RUNNER_IMAGE:
+                c.image, ok = correct_runner_image(c.image)
+                if not ok and ":Host" in c.image:
+                    msg = (
+                        f"Runner image correction failed for Container image {c.image}"
+                    )
+                    raise ValueError(msg)
+
+    def to_json(self) -> str:
+        """
+        Convert the workload plan to a JSON string.
+
+        Returns:
+            The JSON string.
+
+        """
+        return safe_json(self, indent=2)
+
+    def to_yaml(self) -> str:
+        """
+        Convert the workload plan to a YAML string.
+
+        Returns:
+            The YAML string.
+
+        """
+        return safe_yaml(self, indent=2, sort_keys=False)
 
 
 class WorkloadStatusStateEnum(str, Enum):
@@ -885,6 +1049,9 @@ class WorkloadStatusStateEnum(str, Enum):
     The workload is inactive.
     """
 
+    def __str__(self):
+        return self.value
+
 
 WorkloadOperationToken = str
 """
@@ -908,6 +1075,32 @@ class WorkloadStatusOperation:
     Token of the operation, e.g, container ID.
     """
 
+    ##
+    ## The below are internal use only fields.
+    ##
+
+    _name_rfc1123_guard: WorkloadName | None = field(
+        default=None,
+        init=False,
+        repr=False,
+        metadata={"dataclasses_json": {"exclude": lambda _: True}},
+    )
+    """
+    Name for the workload,
+    but has been guard by rfc1123 validation for certain deployment environments,
+    internal use only.
+    """
+
+    @property
+    def name_rfc1123_guard(self) -> WorkloadName:
+        if not self._name_rfc1123_guard:
+            self._name_rfc1123_guard = (
+                f"gpustack-{fnv1a_64_hex(self.name)}"
+                if not is_rfc1123_domain_name(self.name)
+                else self.name
+            )
+        return self._name_rfc1123_guard
+
 
 @dataclass_json
 @dataclass
@@ -1001,37 +1194,36 @@ class Deployer(ABC):
     Base class for all deployers.
     """
 
-    _runtime_visible_devices_env_name: str | None = None
+    _name: str = "unknown"
     """
-    Recorded backend visible devices env name,
-    such as "NVIDIA_VISIBLE_DEVICES", "AMD_VISIBLE_DEVICES", etc.
-    If failed to detect backend, it will be "UNKNOWN_VISIBLE_DEVICES".
+    Name of the deployer.
     """
-    _backend_visible_devices_env_names: list[str] | None = None
+    _pool: ThreadPoolExecutor | None = None
     """
-    Recorded runtime visible devices env name list,
-    such as ["CUDA_VISIBLE_DEVICES"], ["ROCR_VISIBLE_DEVICES"], etc.
-    If failed to detect backend, it will be ["UNKNOWN_VISIBLE_DEVICES"].
+    Thread pool for the deployer.
+    """
+    _visible_devices_env: dict[str, list[str]] | None = None
+    """
+    Recorded visible devices envs,
+    the key is the runtime visible devices env name,
+    the value is the list of backend visible devices env names.
+    For example:
+    {
+        "NVIDIA_VISIBLE_DEVICES": ["CUDA_VISIBLE_DEVICES"],
+        "AMD_VISIBLE_DEVICES": ["HIP_VISIBLE_DEVICES", "ROCR_VISIBLE_DEVICES"]
+    }.
+    """
+    _visible_devices_values: dict[str, list[str]] | None = None
+    """
+    Recorded visible devices values,
+    the key is the runtime visible devices env name,
+    the value is the list of device indexes or uuids.
+    For example:
+    {
+        "NVIDIA_VISIBLE_DEVICES": ["0"],
+        "AMD_VISIBLE_DEVICES": ["0", "1"]
+    }.
     """
-
-    def __init__(self):
-        self._runtime_visible_devices_env_name = (
-            "UNKNOWN_RUNTIME_BACKEND_VISIBLE_DEVICES"
-        )
-        self._backend_visible_devices_env_names = []
-
-        if backend := detect_backend():
-            rk = envs.GPUSTACK_RUNTIME_DETECT_BACKEND_MAP_RESOURCE_KEY.get(backend)
-            ren = envs.GPUSTACK_RUNTIME_DEPLOY_RESOURCE_KEY_MAP_RUNTIME_VISIBLE_DEVICES.get(
-                rk,
-            )
-            ben = envs.GPUSTACK_RUNTIME_DEPLOY_RESOURCE_KEY_MAP_BACKEND_VISIBLE_DEVICES.get(
-                rk,
-            )
-            if ren:
-                self._runtime_visible_devices_env_name = ren
-            if ben:
-                self._backend_visible_devices_env_names = ben
 
     @staticmethod
     @abstractmethod
@@ -1045,8 +1237,177 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
+    @staticmethod
+    def _default_args(func):
+        def wrapper(self, *args, async_mode=None, **kwargs):
+            if async_mode is None:
+                async_mode = envs.GPUSTACK_RUNTIME_DEPLOY_ASYNC
+            return func(
+                self,
+                *args,
+                async_mode=async_mode,
+                **kwargs,
+            )
+
+        return wrapper
+
+    def __init__(self, name: str):
+        self._name = name
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.close()
+
+    def _fetch_visible_devices_env_values(self):
+        """
+        Fetch the visible devices environment variables and values.
+        """
+        if self._visible_devices_env:
+            return
+
+        self._visible_devices_env = {}
+        self._visible_devices_values = {}
+
+        devices: dict[ManufacturerEnum, Devices] = {}
+        for dev in detect_devices(fast=False):
+            if dev.manufacturer not in devices:
+                devices[dev.manufacturer] = []
+            devices[dev.manufacturer].append(dev)
+
+        if devices:
+            value_with_index = (
+                envs.GPUSTACK_RUNTIME_DEPLOY_RUNTIME_VISIBLE_DEVICES_VALUE_MODE.lower()
+                == "index"
+            )
+
+            for manu, devs in devices.items():
+                backend = manufacturer_to_backend(manu)
+                rk = envs.GPUSTACK_RUNTIME_DETECT_BACKEND_MAP_RESOURCE_KEY.get(backend)
+                ren = envs.GPUSTACK_RUNTIME_DEPLOY_RESOURCE_KEY_MAP_RUNTIME_VISIBLE_DEVICES.get(
+                    rk,
+                )
+                ben = envs.GPUSTACK_RUNTIME_DEPLOY_RESOURCE_KEY_MAP_BACKEND_VISIBLE_DEVICES.get(
+                    rk,
+                )
+                if ren and ben:
+                    self._visible_devices_env[ren] = ben
+                    self._visible_devices_values[ren] = [
+                        (str(dev.index) if value_with_index else dev.uuid)
+                        for dev in devs
+                    ]
+
+            if self._visible_devices_env:
+                return
+
+        # Fallback to unknown backend
+        self._visible_devices_env["UNKNOWN_RUNTIME_VISIBLE_DEVICES"] = []
+        self._visible_devices_values["UNKNOWN_RUNTIME_VISIBLE_DEVICES"] = ["all"]
+
+    def visible_devices_env_values(
+        self,
+    ) -> (dict[str, list[str]], dict[str, list[str]]):
+        """
+        Return the visible devices environment variables and values mappings.
+        For example:
+        (
+            {
+                "NVIDIA_VISIBLE_DEVICES": ["CUDA_VISIBLE_DEVICES"],
+                "AMD_VISIBLE_DEVICES": ["HIP_VISIBLE_DEVICES", "ROCR_VISIBLE_DEVICES"]
+            }.
+            {
+                "NVIDIA_VISIBLE_DEVICES": ["0"],
+                "AMD_VISIBLE_DEVICES": ["0", "1"]
+            }
+        ).
+
+        Returns:
+            A tuple of two dictionaries:
+            - The first dictionary maps runtime visible devices environment variable names
+              to lists of backend visible devices environment variable names.
+            - The second dictionary maps runtime visible devices environment variable names
+              to lists of device indexes or UUIDs.
+
+        """
+        self._fetch_visible_devices_env_values()
+        return self._visible_devices_env, self._visible_devices_values
+
+    @property
+    def name(self) -> str:
+        """
+        Return the name of the deployer.
+
+        Returns:
+            The name of the deployer.
+
+        """
+        return self._name
+
+    def close(self):
+        if self._pool:
+            self._pool.shutdown(cancel_futures=True)
+            self._pool = None
+            if hasattr(atexit, "unregister"):
+                atexit.unregister(self.close)
+
+    @property
+    def pool(self):
+        """
+        Create thread pool on first request
+        avoids instantiating unused threadpool for blocking clients.
+        """
+        if self._pool is None:
+            atexit.register(self.close)
+            pool_threads = envs.GPUSTACK_RUNTIME_DEPLOY_ASYNC_THREADS
+            if pool_threads and pool_threads < 1:
+                pool_threads = None
+            self._pool = ThreadPoolExecutor(max_workers=pool_threads)
+        return self._pool
+
+    @_default_args
+    def create(
+        self,
+        workload: WorkloadPlan,
+        async_mode: bool | None = None,
+    ):
+        """
+        Deploy the given workload.
+
+        Args:
+            workload:
+                The workload to deploy.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Raises:
+            TypeError:
+                If the workload type is invalid.
+            ValueError:
+                If the workload fails to validate.
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to deploy.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._create,
+                    workload,
+                )
+                future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload creation failed."
+                raise OperationError(msg) from e
+        else:
+            return self._create(workload)
+
     @abstractmethod
-    def create(self, workload: WorkloadPlan):
+    def _create(self, workload: WorkloadPlan):
         """
         Deploy the given workload.
 
@@ -1067,11 +1428,55 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
-    @abstractmethod
+    @_default_args
     def get(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
+        async_mode: bool | None = None,
+    ) -> WorkloadStatus | None:
+        """
+        Get the status of a workload.
+
+        Args:
+            name:
+                The name of the workload.
+            namespace:
+                The namespace of the workload.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Returns:
+            The status if found, None otherwise.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to get.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._get,
+                    name,
+                    namespace,
+                )
+                return future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload get failed."
+                raise OperationError(msg) from e
+        else:
+            return self._get(name, namespace)
+
+    @abstractmethod
+    def _get(
+        self,
+        name: WorkloadName,
+        namespace: WorkloadNamespace | None = None,
     ) -> WorkloadStatus | None:
         """
         Get the status of a workload.
@@ -1094,11 +1499,55 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
-    @abstractmethod
+    @_default_args
     def delete(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
+        async_mode: bool | None = None,
+    ) -> WorkloadStatus | None:
+        """
+        Delete a workload.
+
+        Args:
+            name:
+                The name of the workload.
+            namespace:
+                The namespace of the workload.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Return:
+            The status if found, None otherwise.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to delete.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._delete,
+                    name,
+                    namespace,
+                )
+                return future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload delete failed."
+                raise OperationError(msg) from e
+        else:
+            return self._delete(name, namespace)
+
+    @abstractmethod
+    def _delete(
+        self,
+        name: WorkloadName,
+        namespace: WorkloadNamespace | None = None,
     ) -> WorkloadStatus | None:
         """
         Delete a workload.
@@ -1121,11 +1570,55 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
-    @abstractmethod
+    @_default_args
     def list(
         self,
         namespace: WorkloadNamespace | None = None,
         labels: dict[str, str] | None = None,
+        async_mode: bool | None = None,
+    ) -> list[WorkloadStatus]:
+        """
+        List all workloads.
+
+        Args:
+            namespace:
+                The namespace of the workloads.
+            labels:
+                Labels to filter the workloads.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Returns:
+            A list of workload statuses.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workloads fail to list.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._list,
+                    namespace,
+                    labels,
+                )
+                return future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload list failed."
+                raise OperationError(msg) from e
+        else:
+            return self._list(namespace, labels)
+
+    @abstractmethod
+    def _list(
+        self,
+        namespace: WorkloadNamespace | None = None,
+        labels: dict[str, str] | None = None,
     ) -> list[WorkloadStatus]:
         """
         List all workloads.
@@ -1148,7 +1641,75 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
-    @abstractmethod
+    async def async_logs(
+        self,
+        name: WorkloadName,
+        namespace: WorkloadNamespace | None = None,
+        token: WorkloadOperationToken | None = None,
+        timestamps: bool = False,
+        tail: int | None = None,
+        since: int | None = None,
+        follow: bool = False,
+        async_mode: bool | None = None,
+    ) -> AsyncGenerator[bytes | str, None, None] | bytes | str:
+        """
+        Asynchronously get the logs of a workload.
+
+        Args:
+            name:
+                The name of the workload to get logs.
+            namespace:
+                The namespace of the workload.
+            token:
+                The operation token of the workload.
+                If not specified, get logs from the first executable container.
+            timestamps:
+                Show timestamps in the logs.
+            tail:
+                Number of lines to show from the end of the logs.
+            since:
+                Show logs since the given epoch in seconds.
+            follow:
+                Whether to follow the logs.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Returns:
+            The logs as a byte string or a generator yielding byte strings if follow is True.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to get logs.
+
+        """
+
+        def run_sync_gen(f: bool) -> Generator[bytes | str, None, None] | bytes | str:
+            return self.logs(
+                name=name,
+                namespace=namespace,
+                token=token,
+                timestamps=timestamps,
+                tail=tail,
+                since=since,
+                follow=f,
+                async_mode=async_mode,
+            )
+
+        if not follow:
+            return run_sync_gen(False)
+
+        async def async_gen() -> AsyncGenerator[bytes | str, None, None]:
+            loop = asyncio.get_event_loop()
+            sync_gen = await loop.run_in_executor(self.pool, run_sync_gen, True)
+            with contextlib.closing(sync_gen) as gen:
+                for chunk in gen:
+                    yield chunk
+
+        return async_gen()
+
+    @_default_args
     def logs(
         self,
         name: WorkloadName,
@@ -1158,6 +1719,71 @@ class Deployer(ABC):
         tail: int | None = None,
         since: int | None = None,
         follow: bool = False,
+        async_mode: bool | None = None,
+    ) -> Generator[bytes | str, None, None] | bytes | str:
+        """
+        Get the logs of a workload.
+
+        Args:
+            name:
+                The name of the workload to get logs.
+            namespace:
+                The namespace of the workload.
+            token:
+                The operation token of the workload.
+                If not specified, get logs from the first executable container.
+            timestamps:
+                Show timestamps in the logs.
+            tail:
+                Number of lines to show from the end of the logs.
+            since:
+                Show logs since the given epoch in seconds.
+            follow:
+                Whether to follow the logs.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Returns:
+            The logs as a byte string or a generator yielding byte strings if follow is True.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to get logs.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._logs,
+                    name,
+                    namespace,
+                    token,
+                    timestamps,
+                    tail,
+                    since,
+                    follow,
+                )
+                return future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload logs failed."
+                raise OperationError(msg) from e
+        else:
+            return self._logs(name, namespace, token, timestamps, tail, since, follow)
+
+    @abstractmethod
+    def _logs(
+        self,
+        name: WorkloadName,
+        namespace: WorkloadNamespace | None = None,
+        token: WorkloadOperationToken | None = None,
+        timestamps: bool = False,
+        tail: int | None = None,
+        since: int | None = None,
+        follow: bool = False,
     ) -> Generator[bytes | str, None, None] | bytes | str:
         """
         Get the logs of a workload.
@@ -1191,7 +1817,7 @@ class Deployer(ABC):
         """
         raise NotImplementedError
 
-    @abstractmethod
+    @_default_args
     def exec(
         self,
         name: WorkloadName,
@@ -1200,6 +1826,69 @@ class Deployer(ABC):
         detach: bool = True,
         command: list[str] | None = None,
         args: list[str] | None = None,
+        async_mode: bool | None = None,
+    ) -> WorkloadExecStream | bytes | str:
+        """
+        Execute a command in a workload.
+
+        Args:
+            name:
+                The name of the workload to execute the command in.
+            namespace:
+                The namespace of the workload.
+            token:
+                The operation token of the workload.
+                If not specified, execute in the first executable container.
+            detach:
+                Whether to detach from the command.
+            command:
+                The command to execute.
+                If not specified, use /bin/sh and implicitly attach.
+            args:
+                The arguments to pass to the command.
+            async_mode:
+                Whether to execute in a separate thread.
+
+        Returns:
+            If detach is False, return a WorkloadExecStream.
+            otherwise, return the output of the command as a byte string or string.
+
+        Raises:
+            UnsupportedError:
+                If the deployer is not supported in the current environment.
+            OperationError:
+                If the workload fails to execute the command.
+
+        """
+        if async_mode:
+            try:
+                future = self.pool.submit(
+                    self._exec,
+                    name,
+                    namespace,
+                    token,
+                    detach,
+                    command,
+                    args,
+                )
+                return future.result()
+            except OperationError:
+                raise
+            except Exception as e:
+                msg = "Asynchronous workload exec failed."
+                raise OperationError(msg) from e
+        else:
+            return self._exec(name, namespace, token, detach, command, args)
+
+    @abstractmethod
+    def _exec(
+        self,
+        name: WorkloadName,
+        namespace: WorkloadNamespace | None = None,
+        token: WorkloadOperationToken | None = None,
+        detach: bool = True,
+        command: list[str] | None = None,
+        args: list[str] | None = None,
     ) -> WorkloadExecStream | bytes | str:
         """
         Execute a command in a workload.
