@@ -1,11 +1,14 @@
 from __future__ import annotations
 
 import contextlib
-import hashlib
 import logging
+import operator
+import os
+import socket
 from dataclasses import dataclass, field
 from enum import Enum
-from functools import lru_cache
+from functools import lru_cache, reduce
+from pathlib import Path
 from typing import TYPE_CHECKING
 
 import kubernetes
@@ -30,14 +33,18 @@ from .__types__ import (
     WorkloadStatusOperation,
     WorkloadStatusStateEnum,
 )
+from .__utils__ import (
+    fnv1a_32_hex,
+    validate_rfc1123_domain_name,
+)
 
 if TYPE_CHECKING:
-    from collections.abc import Generator
-
+    from collections.abc import Callable, Generator
 
 logger = logging.getLogger(__name__)
 
 _LABEL_WORKLOAD = f"{envs.GPUSTACK_RUNTIME_DEPLOY_LABEL_PREFIX}/workload"
+_LABEL_COMPONENT = f"{envs.GPUSTACK_RUNTIME_DEPLOY_LABEL_PREFIX}/component"
 
 
 class KubernetesWorkloadServiceType(str, Enum):
@@ -58,6 +65,9 @@ class KubernetesWorkloadServiceType(str, Enum):
     LoadBalancer: Exposes the service externally using a cloud provider's load balancer.
     """
 
+    def __str__(self):
+        return self.value
+
 
 @dataclass_json
 @dataclass
@@ -121,6 +131,34 @@ class KubernetesWorkloadPlan(WorkloadPlan):
     Service type for the workload.
     """
 
+    def validate_and_default(self):
+        """
+        Validate and set defaults for the workload plan.
+
+        Raises:
+            ValueError:
+                If the workload plan is invalid.
+
+        """
+        if self.labels is None:
+            self.labels = {}
+        if self.containers is None:
+            self.containers = []
+
+        self.labels[_LABEL_WORKLOAD] = self.name
+
+        # Default and validate in the base class.
+        super().validate_and_default()
+
+        # Validate namespace
+        if not self.namespace:
+            self.namespace = envs.GPUSTACK_RUNTIME_KUBERNETES_NAMESPACE
+        try:
+            validate_rfc1123_domain_name(self.namespace)
+        except ValueError as e:
+            msg = f"Invalid namespace '{self.namespace}'"
+            raise ValueError(msg) from e
+
 
 @dataclass_json
 @dataclass
@@ -188,11 +226,6 @@ class KubernetesWorkloadStatus(WorkloadStatus):
         **kwargs,
     ):
         created_at = k_pod.metadata.creation_timestamp.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
-        namespace = (
-            k_pod.metadata.namespace
-            if k_pod.metadata.namespace != envs.GPUSTACK_RUNTIME_KUBERNETES_NAMESPACE
-            else None
-        )
         labels = {
             k: v
             for k, v in (k_pod.metadata.labels or {}).items()
@@ -202,25 +235,28 @@ class KubernetesWorkloadStatus(WorkloadStatus):
         super().__init__(
             name=name,
             created_at=created_at,
-            namespace=namespace,
+            namespace=k_pod.metadata.namespace,
             labels=labels,
             **kwargs,
         )
 
         self._k_pod = k_pod
 
-        for c in k_pod.spec.init_containers or []:
+        k_pod_annos = k_pod.metadata.annotations or {}
+        for ci, c in enumerate(k_pod.spec.init_containers or []):
+            cn = k_pod_annos.get(f"{_LABEL_COMPONENT}-init-{ci}-name", c.name)
             op = WorkloadStatusOperation(
-                name=c.name,
+                name=cn,
                 token=c.name,
             )
             if c.restart_policy != "Never":
                 self.executable.append(op)
             self.loggable.append(op)
 
-        for c in k_pod.spec.containers:
+        for ci, c in enumerate(k_pod.spec.containers):
+            cn = k_pod_annos.get(f"{_LABEL_COMPONENT}-run-{ci}-name", c.name)
             op = WorkloadStatusOperation(
-                name=c.name,
+                name=cn,
                 token=c.name,
             )
             self.executable.append(op)
@@ -229,6 +265,12 @@ class KubernetesWorkloadStatus(WorkloadStatus):
         self.state = self.parse_state(k_pod)
 
 
+_NAME = "kubernetes"
+"""
+Name of the Kubernetes deployer.
+"""
+
+
 class KubernetesDeployer(Deployer):
     """
     Deployer implementation for Kubernetes.
@@ -242,6 +284,12 @@ class KubernetesDeployer(Deployer):
     """
     Name of the node where the deployer is running.
     """
+    _mutate_create_pod: (
+        Callable[[kubernetes.client.V1Pod], kubernetes.client.V1Pod] | None
+    ) = None
+    """
+    Function to handle mirrored deployment, internal use only.
+    """
 
     @staticmethod
     @lru_cache
@@ -254,7 +302,7 @@ class KubernetesDeployer(Deployer):
 
         """
         supported = False
-        if envs.GPUSTACK_RUNTIME_DEPLOY.lower() not in ("auto", "kubernetes"):
+        if envs.GPUSTACK_RUNTIME_DEPLOY.lower() not in ("auto", _NAME):
             return supported
 
         client = KubernetesDeployer._get_client()
@@ -263,10 +311,10 @@ class KubernetesDeployer(Deployer):
                 version_api = kubernetes.client.VersionApi(client)
                 version_info = version_api.get_code()
                 supported = version_info is not None
-            except kubernetes.client.exceptions.ApiException:
-                if logger.isEnabledFor(logging.DEBUG):
-                    logger.exception("Failed to get Kubernetes version")
-            except urllib3.exceptions.MaxRetryError:
+            except (
+                urllib3.exceptions.MaxRetryError,
+                kubernetes.client.exceptions.ApiException,
+            ):
                 if logger.isEnabledFor(logging.DEBUG):
                     logger.exception("Failed to connect to Kubernetes API server")
 
@@ -284,9 +332,14 @@ class KubernetesDeployer(Deployer):
         client = None
 
         try:
-            kubernetes.config.load_kube_config()
-            client = kubernetes.client.ApiClient()
-            client.user_agent = "gpustack/runtime"
+            with (
+                Path(os.devnull).open("w") as dev_null,
+                contextlib.redirect_stdout(dev_null),
+                contextlib.redirect_stderr(dev_null),
+            ):
+                kubernetes.config.load_config()
+                client = kubernetes.client.ApiClient()
+                client.user_agent = "gpustack/runtime"
         except kubernetes.config.config_exception.ConfigException:
             if logger.isEnabledFor(logging.DEBUG):
                 logger.exception("Failed to get Kubernetes client")
@@ -322,23 +375,25 @@ class KubernetesDeployer(Deployer):
                 If creating the ConfigMaps fails.
 
         """
+        config_map_name_prefix = workload.name_rfc1123_guard
+
         ephemeral_filename_mapping: dict[tuple[int, str], str] = {}
         ephemeral_files: list[tuple[str, str, str]] = []
         for ci, c in enumerate(workload.containers):
             for fi, f in enumerate(c.files or []):
                 if f.content is not None:
-                    fn = f"{workload.name}-{ci}-{fi}"
-                    ephemeral_filename_mapping[(ci, f.path)] = fn
-                    ephemeral_files.append((fn, f.path, f.content))
+                    config_map_name = f"{config_map_name_prefix}-{ci}-{fi}"
+                    ephemeral_filename_mapping[(ci, f.path)] = config_map_name
+                    ephemeral_files.append((config_map_name, f.path, f.content))
         if not ephemeral_filename_mapping:
             return ephemeral_filename_mapping
 
         core_api = kubernetes.client.CoreV1Api(self._client)
         try:
-            for fn, fp, content in ephemeral_files:
+            for config_map_name, fp, content in ephemeral_files:
                 config_map = kubernetes.client.V1ConfigMap(
                     metadata=kubernetes.client.V1ObjectMeta(
-                        name=fn,
+                        name=config_map_name,
                         namespace=workload.namespace,
                         labels=workload.labels,
                     ),
@@ -348,7 +403,7 @@ class KubernetesDeployer(Deployer):
                 actual_config_map = None
                 with contextlib.suppress(kubernetes.client.exceptions.ApiException):
                     actual_config_map = core_api.read_namespaced_config_map(
-                        name=fn,
+                        name=config_map_name,
                         namespace=workload.namespace,
                     )
                 if not actual_config_map:
@@ -358,12 +413,12 @@ class KubernetesDeployer(Deployer):
                     )
                     logger.debug(
                         "Created configmap %s in namespace %s",
-                        fn,
+                        config_map_name,
                         workload.namespace,
                     )
                 elif not equal_config_maps(actual_config_map, config_map):
                     core_api.patch_namespaced_config_map(
-                        name=fn,
+                        name=config_map_name,
                         namespace=workload.namespace,
                         body={
                             "data": config_map.data,
@@ -371,11 +426,11 @@ class KubernetesDeployer(Deployer):
                     )
                     logger.debug(
                         "Updated configmap %s in namespace %s",
-                        fn,
+                        config_map_name,
                         workload.namespace,
                     )
         except kubernetes.client.exceptions.ApiException as e:
-            msg = f"Failed to create ephemeral files for workload {workload.name}"
+            msg = f"Failed to create configmap for workload {workload.name}"
             raise OperationError(msg) from e
 
         return ephemeral_filename_mapping
@@ -396,7 +451,7 @@ class KubernetesDeployer(Deployer):
             for _fi, f in enumerate(c.files or []):
                 if f.content is None and f.path:
                     # Host file, bind directly.
-                    fp_hash = hash_string(f.path)
+                    fp_hash = fnv1a_32_hex(f.path)
                     if fp_hash not in file_volumes:
                         file_volumes[fp_hash] = kubernetes.client.V1Volume(
                             name=f"f-{fp_hash}",
@@ -425,7 +480,7 @@ class KubernetesDeployer(Deployer):
             for m in c.mounts or []:
                 if m.volume is None and m.path:
                     # Host directory, bind directly.
-                    mp_hash = hash_string(m.path)
+                    mp_hash = fnv1a_32_hex(m.path)
                     if mp_hash not in mount_volumes:
                         mount_volumes[mp_hash] = kubernetes.client.V1Volume(
                             name=f"m-{mp_hash}",
@@ -468,7 +523,7 @@ class KubernetesDeployer(Deployer):
             for _fi, f in enumerate(files):
                 if f.content is None and f.path:
                     # Host file, mount directly.
-                    fp_hash = hash_string(f.path)
+                    fp_hash = fnv1a_32_hex(f.path)
                     container.volume_mounts.append(
                         kubernetes.client.V1VolumeMount(
                             name=f"f-{fp_hash}",
@@ -492,7 +547,7 @@ class KubernetesDeployer(Deployer):
             for m in mounts:
                 if m.volume is None and m.path:
                     # Host directory, mount directly.
-                    mp_hash = hash_string(m.path)
+                    mp_hash = fnv1a_32_hex(m.path)
                     container.volume_mounts.append(
                         kubernetes.client.V1VolumeMount(
                             name=f"m-{mp_hash}",
@@ -596,9 +651,11 @@ class KubernetesDeployer(Deployer):
         if not ports:
             return None
 
+        service_name = workload.name_rfc1123_guard
+
         service = kubernetes.client.V1Service(
             metadata=kubernetes.client.V1ObjectMeta(
-                name=workload.name,
+                name=service_name,
                 namespace=workload.namespace,
                 labels=workload.labels,
             ),
@@ -622,7 +679,7 @@ class KubernetesDeployer(Deployer):
             actual_service = None
             with contextlib.suppress(kubernetes.client.exceptions.ApiException):
                 actual_service = core_api.read_namespaced_service(
-                    name=workload.name,
+                    name=service_name,
                     namespace=workload.namespace,
                 )
             if not actual_service:
@@ -632,12 +689,12 @@ class KubernetesDeployer(Deployer):
                 )
                 logger.debug(
                     "Created service %s in namespace %s",
-                    workload.name,
+                    service_name,
                     workload.namespace,
                 )
             elif not equal_services(actual_service, service):
                 service = core_api.patch_namespaced_service(
-                    name=workload.name,
+                    name=service_name,
                     namespace=workload.namespace,
                     body={
                         "spec": service.spec,
@@ -645,7 +702,7 @@ class KubernetesDeployer(Deployer):
                 )
                 logger.debug(
                     "Updated service %s in namespace %s",
-                    workload.name,
+                    service_name,
                     workload.namespace,
                 )
         except kubernetes.client.exceptions.ApiException as e:
@@ -672,11 +729,14 @@ class KubernetesDeployer(Deployer):
                 If creating the Pod fails.
 
         """
+        pod_name = workload.name_rfc1123_guard
+
         pod = kubernetes.client.V1Pod(
             metadata=kubernetes.client.V1ObjectMeta(
-                name=workload.name,
+                name=pod_name,
                 namespace=workload.namespace,
                 labels=workload.labels,
+                annotations={},
             ),
             spec=kubernetes.client.V1PodSpec(
                 containers=[],
@@ -695,7 +755,7 @@ class KubernetesDeployer(Deployer):
                             ),
                         ),
                     ]
-                    if workload.shm_size
+                    if not workload.host_ipc and workload.shm_size
                     else None
                 ),
                 security_context=kubernetes.client.V1PodSecurityContext(
@@ -714,10 +774,23 @@ class KubernetesDeployer(Deployer):
             ),
         )
 
+        cnt_init, cnt_run = -1, -1
         for ci, c in enumerate(workload.containers):
+            # Annotate container info.
+            if c.profile == ContainerProfileEnum.INIT:
+                cnt_init += 1
+                container_annotate_prefix = f"{_LABEL_COMPONENT}-init-{cnt_init}"
+            else:
+                cnt_run += 1
+                container_annotate_prefix = f"{_LABEL_COMPONENT}-run-{cnt_run}"
+            pod.metadata.annotations[f"{container_annotate_prefix}-name"] = c.name
+
+            container_name = c.name_rfc1123_guard
+
             container = kubernetes.client.V1Container(
+                name=container_name,
                 image=c.image,
-                name=c.name,
+                image_pull_policy=c.image_pull_policy,
             )
 
             if c.profile == ContainerProfileEnum.INIT:
@@ -756,64 +829,79 @@ class KubernetesDeployer(Deployer):
             # Parameterize resources
             if c.resources:
                 resources: dict[str, str] = {}
-                r_k_rem = workload.resource_key_runtime_env_mapping or {}
-                r_k_bem = workload.resource_key_backend_env_mapping or {}
+                r_k_runtime_env = workload.resource_key_runtime_env_mapping or {}
+                r_k_backend_env = workload.resource_key_backend_env_mapping or {}
+                vd_env, vd_values = self.visible_devices_env_values()
                 for r_k, r_v in c.resources.items():
                     if r_k in ("cpu", "memory"):
                         resources[r_k] = str(r_v)
                     else:
-                        re = None
-                        if r_k in r_k_rem:
+                        if r_k in r_k_runtime_env:
                             # Set env if resource key is mapped.
-                            re = r_k_rem[r_k]
+                            runtime_env = r_k_runtime_env[r_k]
                         elif r_k == envs.GPUSTACK_RUNTIME_DEPLOY_AUTOMAP_RESOURCE_KEY:
                             # Set env if auto-mapping key is matched.
-                            re = self._runtime_visible_devices_env_name
-                        if not re:
+                            runtime_env = list(vd_env.keys())
+                        else:
                             resources[r_k] = str(r_v)
                             continue
 
-                        bes = []
-                        if r_k in r_k_bem:
+                        if r_k in r_k_backend_env:
                             # Set env if resource key is mapped.
-                            bes = r_k_bem[r_k]
+                            backend_env = r_k_backend_env[r_k]
                         else:
                             # Otherwise, use the default backend env names.
-                            bes = self._backend_visible_devices_env_names
+                            backend_env = reduce(operator.add, list(vd_env.values()))
+
+                        privileged = (
+                            container.security_context
+                            and container.security_context.privileged
+                        )
 
                         # Configure device access environment variable.
-                        if r_v == "all" and bes:
+                        if r_v == "all" and backend_env:
                             # Configure privileged if requested all devices.
                             container.security_context = (
                                 container.security_context
                                 or kubernetes.client.V1SecurityContext()
                             )
                             container.security_context.privileged = True
-                            # Then, set container backend visible devices env to "0",
+                            # Then, set container backend visible devices env to all devices,
                             # so that the container backend (e.g., NVIDIA Container Toolkit) can handle it,
                             # and mount corresponding libs if needed.
-                            container.env.append(
-                                kubernetes.client.V1EnvVar(
-                                    name=re,
-                                    value="0",
-                                ),
-                            )
+                            for re in runtime_env:
+                                # Set to "all" if no specific devices detected,
+                                # maybe the container backend can handle it.
+                                container.env.append(
+                                    kubernetes.client.V1EnvVar(
+                                        name=re,
+                                        value=",".join(vd_values.get(re, [])) or "all",
+                                    ),
+                                )
                         else:
-                            # Set env to the allocated device IDs.
-                            container.env.append(
-                                kubernetes.client.V1EnvVar(
-                                    name=re,
-                                    value=str(r_v),
-                                ),
-                            )
+                            # Set env to the allocated device IDs if no privileged,
+                            # otherwise, set container backend visible devices env to all devices,
+                            # so that the container backend (e.g., NVIDIA Container Toolkit) can handle it,
+                            # and mount corresponding libs if needed.
+                            for re in runtime_env:
+                                # Set to "all" if no specific devices detected,
+                                # maybe the container backend can handle it.
+                                container.env.append(
+                                    kubernetes.client.V1EnvVar(
+                                        name=re,
+                                        value=(
+                                            r_v
+                                            if not privileged
+                                            else (
+                                                ",".join(vd_values.get(re, [])) or "all"
+                                            )
+                                        ),
+                                    ),
+                                )
 
                         # Configure runtime device access environment variables.
-                        if (
-                            r_v != "all"
-                            and container.security_context
-                            and container.security_context.privileged
-                        ):
-                            for be in bes:
+                        if r_v != "all" and privileged:
+                            for be in backend_env:
                                 container.env.append(
                                     kubernetes.client.V1EnvVar(
                                         name=be,
@@ -875,7 +963,7 @@ class KubernetesDeployer(Deployer):
                     container.readiness_probe = self._parameterize_probe(chk)
 
             # Concat shared memory volume if needed.
-            if workload.shm_size:
+            if not workload.host_ipc and workload.shm_size:
                 if not container.volume_mounts:
                     container.volume_mounts = []
                 container.volume_mounts.append(
@@ -892,12 +980,14 @@ class KubernetesDeployer(Deployer):
             else:
                 pod.spec.containers.append(container)
 
+        pod = self._mutate_create_pod(pod)
+
         core_api = kubernetes.client.CoreV1Api(self._client)
         try:
             actual_pod = None
             with contextlib.suppress(kubernetes.client.exceptions.ApiException):
                 actual_pod = core_api.read_namespaced_pod(
-                    name=workload.name,
+                    name=pod_name,
                     namespace=workload.namespace,
                 )
             if not actual_pod:
@@ -907,7 +997,7 @@ class KubernetesDeployer(Deployer):
                 )
                 logger.debug(
                     "Created pod %s in namespace %s",
-                    workload.name,
+                    pod_name,
                     workload.namespace,
                 )
             elif not equal_pods(actual_pod, pod):
@@ -920,13 +1010,13 @@ class KubernetesDeployer(Deployer):
                     namespace=workload.namespace,
                 ) as es:
                     core_api.delete_namespaced_pod(
-                        name=workload.name,
+                        name=pod_name,
                         namespace=workload.namespace,
                     )
                     for e in es:
                         if (
                             e["type"] == "DELETED"
-                            and e["object"].metadata.name == workload.name
+                            and e["object"].metadata.name == pod_name
                         ):
                             break
                 pod = core_api.create_namespaced_pod(
@@ -935,7 +1025,7 @@ class KubernetesDeployer(Deployer):
                 )
                 logger.debug(
                     "Updated pod %s in namespace %s",
-                    workload.name,
+                    pod_name,
                     workload.namespace,
                 )
         except kubernetes.client.exceptions.ApiException as e:
@@ -945,20 +1035,251 @@ class KubernetesDeployer(Deployer):
         return pod
 
     def __init__(self):
-        super().__init__()
+        super().__init__(_NAME)
         self._client = self._get_client()
         self._node_name = envs.GPUSTACK_RUNTIME_KUBERNETES_NODE_NAME
-        if self._client and not self._node_name:
-            # Get the first node name of the cluster if not configured.
-            # Required list permission on Kubernetes Node resources.
+
+    def _prepare_create(self):
+        """
+        Prepare for creation.
+
+        """
+        # Get the first node name of the cluster if not configured.
+        # Required list permission on Kubernetes Node resources.
+        if not self._node_name:
             core_api = kubernetes.client.CoreV1Api(self._client)
-            with contextlib.suppress(kubernetes.client.exceptions.ApiException):
+            try:
                 nodes = core_api.list_node(limit=1)
+            except kubernetes.client.exceptions.ApiException as e:
+                msg = "Failed to get the default node name of the cluster"
+                raise OperationError(msg) from e
+            else:
                 if nodes.items:
                     self._node_name = nodes.items[0].metadata.name
+                else:
+                    msg = "Failed to get the default node name of the cluster: No nodes found"
+                    raise OperationError(msg)
+
+        # Prepare mirrored deployment if enabled.
+        if self._mutate_create_pod:
+            return
+        self._mutate_create_pod = lambda o: o
+        if not envs.GPUSTACK_RUNTIME_DEPLOY_MIRRORED_DEPLOYMENT:
+            logger.debug("Mirrored deployment disabled")
+            return
+
+        # Retrieve self-pod info.
+        core_api = kubernetes.client.CoreV1Api(self._client)
+        ## - Get Pod name, default to hostname if not set.
+        self_pod_name = envs.GPUSTACK_RUNTIME_DEPLOY_MIRRORED_NAME
+        if not self_pod_name:
+            self_pod_name = socket.gethostname()
+            logger.warning(
+                "Mirrored deployment enabled, but no Pod name set, using hostname(%s) instead",
+                self_pod_name,
+            )
+        ## - Get Pod namespace, default to "default" if not found.
+        try:
+            self_pod_namespace_f = Path(
+                "/var/run/secrets/kubernetes.io/serviceaccount/namespace",
+            )
+            self_pod_namespace = self_pod_namespace_f.read_text(
+                encoding="utf-8",
+            ).strip()
+        except (FileNotFoundError, OSError):
+            self_pod_namespace = "default"
+            logger.warning(
+                "Mirrored deployment enabled, but no Pod namespace found, using 'default' instead",
+            )
+        try:
+            self_pod = core_api.read_namespaced_pod(
+                name=self_pod_name,
+                namespace=self_pod_namespace,
+            )
+        except kubernetes.client.exceptions.ApiException:
+            output_log = logger.warning
+            if logger.isEnabledFor(logging.DEBUG):
+                output_log = logger.exception
+            output_log(
+                f"Mirrored deployment enabled, but failed to get self Pod {self_pod_namespace}/{self_pod_name}, skipping",
+            )
+            return
+        ## - Get the first Container, or the Container named "default" if exists.
+        self_container = next(
+            (c for c in self_pod.spec.containers if c.name == "default"),
+            None,
+        )
+        if not self_container:
+            self_container = self_pod.spec.containers[0]
+            logger.warning(
+                "Mirrored deployment enabled, but no Container named 'default' found, using the first Container instead",
+            )
+
+        # Preprocess mirrored deployment options.
+        in_same_namespace = (
+            self_pod_namespace == envs.GPUSTACK_RUNTIME_KUBERNETES_NAMESPACE
+        )
+        ## - Pod runtime class name
+        mirrored_runtime_class_name: str = self_pod.spec.runtime_class_name or ""
+        ## - Container envs
+        mirrored_envs: list[kubernetes.client.V1EnvVar] = [
+            # Filter out gpustack-internal envs and cross-namespace secret/envref envs.
+            e
+            for e in self_container.env or []
+            if (
+                not e.name.startswith("GPUSTACK_")
+                and (not e.value_from or in_same_namespace)
+            )
+        ]
+        if igs := envs.GPUSTACK_RUNTIME_DEPLOY_MIRRORED_DEPLOYMENT_IGNORE_ENVIRONMENTS:
+            mirrored_envs = [
+                # Filter out ignored envs.
+                e
+                for e in mirrored_envs
+                if e.name not in igs
+            ]
+        ## - Container volume mounts
+        mirrored_volume_mounts: list[kubernetes.client.V1VolumeMount] = (
+            self_container.volume_mounts or []
+        )
+        if igs := envs.GPUSTACK_RUNTIME_DEPLOY_MIRRORED_DEPLOYMENT_IGNORE_VOLUMES:
+            mirrored_volume_mounts = [
+                # Filter out ignored volume mounts.
+                m
+                for m in mirrored_volume_mounts
+                if m.mount_path not in igs
+            ]
+        ## - Container volume devices
+        mirrored_volume_devices: list[kubernetes.client.V1VolumeDevice] = (
+            self_container.volume_devices or []
+        )
+        if igs := envs.GPUSTACK_RUNTIME_DEPLOY_MIRRORED_DEPLOYMENT_IGNORE_VOLUMES:
+            mirrored_volume_devices = [
+                # Filter out ignored volume mounts.
+                d
+                for d in mirrored_volume_devices
+                if d.device_path not in igs
+            ]
+        ## - Pod volumes
+        mirrored_volume_mounts_names = {m.name for m in mirrored_volume_mounts}
+        mirrored_volume_devices_names = {d.name for d in mirrored_volume_devices}
+        mirrored_volumes: list[kubernetes.client.V1Volume] = []
+        # Filter out volumes not used by mirrored volume mounts or devices.
+        for v in self_pod.spec.volumes or []:
+            if (
+                v.name not in mirrored_volume_mounts_names
+                and v.name not in mirrored_volume_devices_names
+            ):
+                continue
+            # Skip downwardAPI/projected volumes
+            if v.downward_api or v.projected:
+                mirrored_volume_mounts_names.discard(v.name)
+                mirrored_volume_devices_names.discard(v.name)
+                continue
+            # Skip configMap/secret/PVC volumes if not in same namespace
+            if (
+                v.config_map or v.secret or v.persistent_volume_claim
+            ) and not in_same_namespace:
+                mirrored_volume_mounts_names.discard(v.name)
+                mirrored_volume_devices_names.discard(v.name)
+                continue
+            # Skip PVCs with RWO or RWOP access modes
+            if v.persistent_volume_claim:
+                pvc = core_api.read_namespaced_persistent_volume_claim(
+                    name=v.persistent_volume_claim.claim_name,
+                    namespace=self_pod_namespace,
+                )
+                ams = pvc.spec.access_modes or []
+                if any(mode in ams for mode in ("ReadWriteOnce", "ReadWriteOncePod")):
+                    mirrored_volume_mounts_names.discard(v.name)
+                    mirrored_volume_devices_names.discard(v.name)
+                    continue
+            mirrored_volumes.append(v)
+        ## - Correct Container volume mounts and volume devices without corresponding volumes.
+        mirrored_volume_mounts = [
+            m for m in mirrored_volume_mounts if m.name in mirrored_volume_mounts_names
+        ]
+        mirrored_volume_devices = [
+            d
+            for d in mirrored_volume_devices
+            if d.name in mirrored_volume_devices_names
+        ]
+
+        # Construct mutation function.
+        def mutate_create_pod(
+            pod: kubernetes.client.V1Pod,
+        ) -> kubernetes.client.V1Pod:
+            if mirrored_runtime_class_name and not pod.spec.runtime_class_name:
+                pod.spec.runtime_class_name = mirrored_runtime_class_name
+
+            if mirrored_envs:
+                for ci in pod.spec.containers:
+                    c_env_names = {ei.name for ei in ci.env or []}
+                    for ei in mirrored_envs:
+                        if ei.name not in c_env_names:
+                            ci.env.append(ei)
+
+            if mirrored_volume_mounts or mirrored_volume_devices:
+                for ci in pod.spec.containers:
+                    c_volume_mount_names = {
+                        # Map existing volume mount names.
+                        mi.name
+                        for mi in ci.volume_mounts or []
+                    }
+                    c_volume_mount_paths = {
+                        # Map existing volume mount paths.
+                        mi.mount_path
+                        for mi in ci.volume_mounts or []
+                    }
+                    # Append volume mounts if not exists.
+                    for mi in mirrored_volume_mounts:
+                        if (
+                            mi.name not in c_volume_mount_names
+                            and mi.mount_path not in c_volume_mount_paths
+                        ):
+                            ci.volume_mounts = ci.volume_mounts or []
+                            ci.volume_mounts.append(mi)
+                            c_volume_mount_names.add(mi.name)
+                            c_volume_mount_paths.add(mi.mount_path)
+                    # Append volume devices if not exists.
+                    c_volume_device_names = {
+                        # Map existing volume device names.
+                        di.name
+                        for di in ci.volume_devices or []
+                    }
+                    c_volume_device_paths = {
+                        # Map existing volume device paths.
+                        di.device_path
+                        for di in ci.volume_devices or []
+                    }
+                    for di in mirrored_volume_devices:
+                        if (
+                            di.name not in c_volume_device_names
+                            and di.device_path not in c_volume_device_paths
+                        ):
+                            ci.volume_devices = ci.volume_devices or []
+                            ci.volume_devices.append(di)
+                            c_volume_device_names.add(di.name)
+                            c_volume_device_paths.add(di.device_path)
+
+            if mirrored_volumes:
+                p_volume_names = {
+                    # Map existing volume names.
+                    vi.name
+                    for vi in pod.spec.volumes or []
+                }
+                for vi in mirrored_volumes:
+                    if vi.name not in p_volume_names:
+                        pod.spec.volumes = pod.spec.volumes or []
+                        pod.spec.volumes.append(vi)
+                        p_volume_names.add(vi.name)
+
+            return pod
+
+        self._mutate_create_pod = mutate_create_pod
 
     @_supported
-    def create(self, workload: WorkloadPlan):
+    def _create(self, workload: WorkloadPlan):
         """
         Deploy a Kubernetes workload.
 
@@ -980,16 +1301,14 @@ class KubernetesDeployer(Deployer):
         if not isinstance(workload, KubernetesWorkloadPlan | WorkloadPlan):
             msg = f"Invalid workload plan type: {type(workload)}"
             raise TypeError(msg)
+
+        self._prepare_create()
+
         if isinstance(workload, WorkloadPlan):
             workload = KubernetesWorkloadPlan(**workload.__dict__)
-
         workload.validate_and_default()
-        workload.labels[_LABEL_WORKLOAD] = workload.name
-        workload.namespace = (
-            workload.namespace or envs.GPUSTACK_RUNTIME_KUBERNETES_NAMESPACE
-        )
         if logger.isEnabledFor(logging.DEBUG):
-            logger.debug("Creating workload:\n%s", workload.to_json(indent=2))
+            logger.debug("Creating workload:\n%s", workload.to_yaml())
 
         # Create ephemeral file if needed,
         # (container index, configured path): <actual ConfigMap name>
@@ -1007,7 +1326,7 @@ class KubernetesDeployer(Deployer):
         )
 
     @_supported
-    def get(
+    def _get(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
@@ -1071,7 +1390,7 @@ class KubernetesDeployer(Deployer):
         )
 
     @_supported
-    def delete(
+    def _delete(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
@@ -1140,7 +1459,7 @@ class KubernetesDeployer(Deployer):
         return workload
 
     @_supported
-    def list(
+    def _list(
         self,
         namespace: WorkloadNamespace | None = None,
         labels: dict[str, str] | None = None,
@@ -1207,7 +1526,7 @@ class KubernetesDeployer(Deployer):
         ]
 
     @_supported
-    def logs(
+    def _logs(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
@@ -1292,7 +1611,7 @@ class KubernetesDeployer(Deployer):
             return output
 
     @_supported
-    def exec(
+    def _exec(
         self,
         name: WorkloadName,
         namespace: WorkloadNamespace | None = None,
@@ -1380,21 +1699,6 @@ class KubernetesDeployer(Deployer):
             return KubernetesWorkloadExecStream(result)
 
 
-def hash_string(s: str) -> str:
-    """
-    Hash a string using MD5 and return the hexadecimal digest.
-
-    Args:
-        s:
-            The string to hash.
-
-    Returns:
-        The hexadecimal digest of the MD5 hash.
-
-    """
-    return hashlib.md5(s.encode("UTF-8")).hexdigest()  # noqa: S324
-
-
 def equal_config_maps(
     a: kubernetes.client.V1ConfigMap,
     b: kubernetes.client.V1ConfigMap,
@@ -1476,6 +1780,8 @@ def equal_containers(
         return False
     if (a.volume_mounts or []) != (b.volume_mounts or []):
         return False
+    if (a.volume_devices or []) != (b.volume_devices or []):
+        return False
     if (a.liveness_probe or {}) != (b.liveness_probe or {}):
         return False
     if (a.readiness_probe or {}) != (b.readiness_probe or {}):
@@ -1517,6 +1823,8 @@ def equal_pods(
     ):
         if not equal_containers(ac, bc):
             return False
+    if aspec.runtime_class_name != bspec.runtime_class_name:
+        return False
     if aspec.host_network != bspec.host_network:
         return False
     if aspec.host_ipc != bspec.host_ipc:
