@@ -33,6 +33,10 @@ class ManufacturerEnum(str, Enum):
     """
     Shanghai Iluvatar CoreX Semiconductor Co., Ltd.
     """
+    METAX = "metax"
+    """
+    MetaX Integrated Circuits (Shanghai) Co., Ltd.
+    """
     MTHREADS = "mthreads"
     """
     Moore Threads Technology Co.,Ltd
@@ -46,13 +50,17 @@ class ManufacturerEnum(str, Enum):
     Unknown Manufacturer
     """
 
+    def __str__(self):
+        return self.value
+
 
 _MANUFACTURER_BACKEND_MAPPING: dict[ManufacturerEnum, str] = {
     ManufacturerEnum.AMD: "rocm",
     ManufacturerEnum.ASCEND: "cann",
-    ManufacturerEnum.CAMBRICON: "cnrt",
+    ManufacturerEnum.CAMBRICON: "neuware",
     ManufacturerEnum.HYGON: "dtk",
     ManufacturerEnum.ILUVATAR: "corex",
+    ManufacturerEnum.METAX: "maca",
     ManufacturerEnum.MTHREADS: "musa",
     ManufacturerEnum.NVIDIA: "cuda",
 }
@@ -62,7 +70,7 @@ which should map to the gpustack-runner's backend names.
 """
 
 
-def manufacturer_to_backend(manufacturer: ManufacturerEnum) -> str | None:
+def manufacturer_to_backend(manufacturer: ManufacturerEnum) -> str:
     """
     Convert manufacturer to runtime backend,
     e.g., NVIDIA -> cuda, AMD -> rocm.
@@ -74,13 +82,17 @@ def manufacturer_to_backend(manufacturer: ManufacturerEnum) -> str | None:
         manufacturer: The manufacturer of the device.
 
     Returns:
-        The corresponding runtime backend. None if the manufacturer is unknown.
+        The corresponding runtime backend.
+        Return "unknown" if the manufacturer is unknown.
 
     """
-    return _MANUFACTURER_BACKEND_MAPPING.get(manufacturer)
+    backend = _MANUFACTURER_BACKEND_MAPPING.get(manufacturer)
+    if backend:
+        return backend
+    return ManufacturerEnum.UNKNOWN.value
 
 
-def backend_to_manufacturer(backend: str) -> ManufacturerEnum | None:
+def backend_to_manufacturer(backend: str) -> ManufacturerEnum:
     """
     Convert runtime backend to manufacturer,
     e.g., cuda -> NVIDIA, rocm -> AMD.
@@ -92,13 +104,14 @@ def backend_to_manufacturer(backend: str) -> ManufacturerEnum | None:
         backend: The runtime backend.
 
     Returns:
-        The corresponding manufacturer. None if the backend is unknown.
+        The corresponding manufacturer.
+        Return ManufacturerEnum.Unknown if the backend is unknown.
 
     """
     for manufacturer, mapped_backend in _MANUFACTURER_BACKEND_MAPPING.items():
         if mapped_backend == backend:
             return manufacturer
-    return None
+    return ManufacturerEnum.UNKNOWN
 
 
 def supported_manufacturers() -> list[ManufacturerEnum]:
@@ -154,29 +167,18 @@ class Device:
     """
     Driver version of the device.
     """
-    driver_version_tuple: list[int | str] | None = None
-    """
-    Driver version tuple of the device.
-    None if `driver_version` is missed.
-    """
     runtime_version: str | None = None
     """
-    Runtime version of the device.
+    Runtime version in major[.minor] of the device.
     """
-    runtime_version_tuple: list[int | str] | None = None
+    runtime_version_original: str | None = None
     """
-    Runtime version tuple of the device.
-    None if `runtime_version` is missed.
+    Original runtime version string of the device.
     """
     compute_capability: str | None = None
     """
     Compute capability of the device.
     """
-    compute_capability_tuple: list[int | str] | None = None
-    """
-    Compute capability tuple of the device.
-    None if `compute_capability` is missed.
-    """
     cores: int | None = None
     """
     Total cores of the device.
@@ -185,15 +187,15 @@ class Device:
     """
     Core utilization of the device in percentage.
     """
-    memory: int | float = 0
+    memory: int = 0
     """
     Total memory of the device in MiB.
     """
-    memory_used: int | float = 0
+    memory_used: int = 0
     """
     Used memory of the device in MiB.
     """
-    memory_utilization: int | float = 0
+    memory_utilization: float = 0
     """
     Memory utilization of the device in percentage.
     """
@@ -244,12 +246,19 @@ class Detector(ABC):
         self.manufacturer = manufacturer
 
     @property
-    def backend(self) -> str | None:
+    def backend(self) -> str:
         """
         The backend name of the detector, e.g., 'cuda', 'rocm'.
         """
         return manufacturer_to_backend(self.manufacturer)
 
+    @property
+    def name(self) -> str:
+        """
+        The name of the detector, e.g., 'nvidia', 'amd'.
+        """
+        return str(self.manufacturer)
+
     @abstractmethod
     def detect(self) -> Devices | None:
         """
