@@ -5,9 +5,14 @@ import logging
 from functools import lru_cache
 
 from .. import envs
-from . import pyamdgpu, pyamdsmi, pyrocmsmi
+from . import pyamdgpu, pyamdsmi, pyhsa, pyrocmcore, pyrocmsmi
 from .__types__ import Detector, Device, Devices, ManufacturerEnum
-from .__utils__ import PCIDevice, get_device_files, get_pci_devices
+from .__utils__ import (
+    PCIDevice,
+    get_brief_version,
+    get_pci_devices,
+    get_utilization,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -77,66 +82,63 @@ class AMDDetector(Detector):
         ret: Devices = []
 
         try:
+            hsa_agents = {hsa_agent.uuid: hsa_agent for hsa_agent in pyhsa.get_agents()}
+
             pyamdsmi.amdsmi_init()
 
-            sys_runtime_ver = pyamdsmi.amdsmi_get_rocm_version2()
-            sys_runtime_ver_t = (
-                [int(v) if v.isdigit() else v for v in sys_runtime_ver.split(".")]
-                if sys_runtime_ver
-                else None
-            )
+            sys_runtime_ver_original = pyrocmcore.getROCmVersion()
+            sys_runtime_ver = get_brief_version(sys_runtime_ver_original)
 
             devs = pyamdsmi.amdsmi_get_processor_handles()
-            dev_files = get_device_files(
-                pattern=r"card(?P<number>\d+)",
-                directory="/dev/dri",
-            )
             for dev_idx, dev in enumerate(devs):
-                dev_card = None
                 dev_index = dev_idx
-                if len(dev_files) > dev_idx:
-                    dev_file = dev_files[dev_idx]
-                    if dev_file.number is not None:
-                        dev_card = dev_file.number
-                        if envs.GPUSTACK_RUNTIME_DETECT_PHYSICAL_INDEX_PRIORITY:
-                            dev_index = (
-                                dev_file.number - 1 if dev_file.number > 0 else 0
-                            )
+
+                dev_gpu_asic_info = pyamdsmi.amdsmi_get_gpu_asic_info(dev)
+                dev_uuid = f"GPU-{(dev_gpu_asic_info.get('asic_serial')[2:]).lower()}"
+                dev_hsa_agent = hsa_agents.get(dev_uuid)
+
+                dev_card_id = None
+                if dev_hsa_agent:
+                    dev_card_id = dev_hsa_agent.driver_node_id
+                elif hasattr(pyamdsmi, "amdsmi_get_gpu_kfd_info"):
+                    dev_kfd_info = pyamdsmi.amdsmi_get_gpu_kfd_info(dev)
+                    dev_card_id = dev_kfd_info.get("node_id")
+                else:
+                    with contextlib.suppress(pyrocmsmi.ROCMSMIError):
+                        pyrocmsmi.rsmi_init()
+                        dev_card_id = pyrocmsmi.rsmi_dev_node_id_get(dev_idx)
 
                 dev_gpudev_info = None
-                if dev_card is not None:
+                if dev_card_id is not None:
                     with contextlib.suppress(pyamdgpu.AMDGPUError):
-                        _, _, dev_gpudev = pyamdgpu.amdgpu_device_initialize(dev_card)
+                        _, _, dev_gpudev = pyamdgpu.amdgpu_device_initialize(
+                            dev_card_id,
+                        )
                         dev_gpudev_info = pyamdgpu.amdgpu_query_gpu_info(dev_gpudev)
                         pyamdgpu.amdgpu_device_deinitialize(dev_gpudev)
 
                 dev_gpu_driver_info = pyamdsmi.amdsmi_get_gpu_driver_info(dev)
                 dev_driver_ver = dev_gpu_driver_info.get("driver_version")
-                dev_driver_ver_t = (
-                    [int(v) if v.isdigit() else v for v in dev_driver_ver.split(".")]
-                    if dev_driver_ver
-                    else None
-                )
 
-                dev_gpu_asic_info = pyamdsmi.amdsmi_get_gpu_asic_info(dev)
-                dev_uuid = dev_gpu_asic_info.get("asic_serial")
-                dev_name = "AMD " + dev_gpu_asic_info.get("market_name")
-                dev_cc = None
-                dev_cc_t = None
-                if hasattr(dev_gpu_asic_info, "target_graphics_version"):
-                    dev_cc = dev_gpu_asic_info.target_graphics_version
+                if dev_hsa_agent:
+                    dev_name = dev_hsa_agent.name
+                    if not dev_name:
+                        dev_name = dev_gpu_asic_info.get("market_name")
+                    dev_cc = dev_hsa_agent.compute_capability
                 else:
+                    dev_name = dev_gpu_asic_info.get("market_name")
+                    dev_cc = None
                     with contextlib.suppress(pyrocmsmi.ROCMSMIError):
                         pyrocmsmi.rsmi_init()
                         dev_cc = pyrocmsmi.rsmi_dev_target_graphics_version_get(dev_idx)
-                if dev_cc:
-                    dev_cc = dev_cc[3:]  # Strip "gfx" prefix
-                    dev_cc_t = [int(v) if v.isdigit() else v for v in dev_cc.split(".")]
+
+                dev_cores = None
+                if dev_hsa_agent:
+                    dev_cores = dev_hsa_agent.compute_units
+                elif dev_gpudev_info and hasattr(dev_gpudev_info, "cu_active_number"):
+                    dev_cores = dev_gpudev_info.cu_active_number
 
                 dev_gpu_metrics_info = pyamdsmi.amdsmi_get_gpu_metrics_info(dev)
-                dev_cores = (
-                    dev_gpudev_info.cu_active_number if dev_gpudev_info else None
-                )
                 dev_cores_util = dev_gpu_metrics_info.get("average_gfx_activity", 0)
                 dev_gpu_vram_usage = pyamdsmi.amdsmi_get_gpu_vram_usage(dev)
                 dev_mem = dev_gpu_vram_usage.get("vram_total")
@@ -160,6 +162,7 @@ class AMDDetector(Detector):
                 dev_appendix = {
                     "arch_family": _get_arch_family(dev_gpudev_info),
                     "vgpu": dev_compute_partition is not None,
+                    "card_id": dev_card_id,
                 }
 
                 ret.append(
@@ -169,18 +172,14 @@ class AMDDetector(Detector):
                         name=dev_name,
                         uuid=dev_uuid,
                         driver_version=dev_driver_ver,
-                        driver_version_tuple=dev_driver_ver_t,
                         runtime_version=sys_runtime_ver,
-                        runtime_version_tuple=sys_runtime_ver_t,
+                        runtime_version_original=sys_runtime_ver_original,
                         compute_capability=dev_cc,
-                        compute_capability_tuple=dev_cc_t,
                         cores=dev_cores,
                         cores_utilization=dev_cores_util,
                         memory=dev_mem,
                         memory_used=dev_mem_used,
-                        memory_utilization=(
-                            (dev_mem_used * 100 // dev_mem) if dev_mem > 0 else 0
-                        ),
+                        memory_utilization=get_utilization(dev_mem_used, dev_mem),
                         temperature=dev_temp,
                         power=dev_power,
                         power_used=dev_power_used,
