@@ -1,14 +1,16 @@
 from __future__ import annotations
 
+import contextlib
 import logging
 from functools import lru_cache
 
 from .. import envs
-from . import pydcmi
+from . import pyacl, pydcmi
 from .__types__ import Detector, Device, Devices, ManufacturerEnum
-from .__utils__ import PCIDevice, get_pci_devices
+from .__utils__ import PCIDevice, get_brief_version, get_pci_devices, get_utilization
 
 logger = logging.getLogger(__name__)
+slogger = logger.getChild("internal")
 
 
 class AscendDetector(Detector):
@@ -78,16 +80,9 @@ class AscendDetector(Detector):
             pydcmi.dcmi_init()
 
             sys_driver_ver = pydcmi.dcmi_get_driver_version()
-            sys_driver_ver_t = [
-                int(v) if v.isdigit() else v for v in sys_driver_ver.split(".")
-            ]
 
-            sys_runtime_ver = pydcmi.dcmi_get_cann_version()
-            sys_runtime_ver_t = (
-                [int(v) if v.isdigit() else v for v in sys_runtime_ver.split(".")]
-                if sys_runtime_ver
-                else None
-            )
+            sys_runtime_ver_original = pyacl.aclsysGetCANNVersion()
+            sys_runtime_ver = get_brief_version(sys_runtime_ver_original)
 
             _, card_list = pydcmi.dcmi_get_card_list()
             for dev_card_id in card_list:
@@ -105,7 +100,7 @@ class AscendDetector(Detector):
                     ):
                         dev_is_vgpu = True
                         dev_cores_aicore = dev_virt_info.query_info.computing.aic
-                        dev_name = f"Ascend {dev_virt_info.query_info.name}"
+                        dev_name = dev_virt_info.query_info.name
                         dev_mem, dev_mem_used = 0, 0
                         if hasattr(dev_virt_info.query_info.computing, "memory_size"):
                             dev_mem = dev_virt_info.query_info.computing.memory_size
@@ -116,9 +111,7 @@ class AscendDetector(Detector):
                             dev_device_id,
                         )
                         dev_cores_aicore = dev_chip_info.aicore_cnt
-                        dev_name = (
-                            f"{dev_chip_info.chip_type} {dev_chip_info.chip_name}"
-                        )
+                        dev_name = dev_chip_info.chip_name
                         dev_mem, dev_mem_used = _get_device_memory_info(
                             dev_card_id,
                             dev_device_id,
@@ -145,11 +138,19 @@ class AscendDetector(Detector):
                         dev_card_id,
                         dev_device_id,
                     )
-                    dev_power_used = pydcmi.dcmi_get_device_power_info(
-                        dev_card_id,
-                        dev_device_id,
-                    )
+                    dev_power_used = None
+                    with contextlib.suppress(pydcmi.DCMIError):
+                        dev_power_used = pydcmi.dcmi_get_device_power_info(
+                            dev_card_id,
+                            dev_device_id,
+                        )
+                    if dev_power_used:
+                        dev_power_used = dev_power_used / 10  # 0.1W to W
                     dev_appendix = {
+                        "arch_family": (
+                            pyacl.aclrtGetSocName()
+                            or _guess_soc_name_from_dev_name(dev_name)
+                        ),
                         "vgpu": dev_is_vgpu,
                         "card_id": dev_card_id,
                         "device_id": dev_device_id,
@@ -174,18 +175,15 @@ class AscendDetector(Detector):
                             name=dev_name,
                             uuid=dev_uuid.upper(),
                             driver_version=sys_driver_ver,
-                            driver_version_tuple=sys_driver_ver_t,
                             runtime_version=sys_runtime_ver,
-                            runtime_version_tuple=sys_runtime_ver_t,
+                            runtime_version_original=sys_runtime_ver_original,
                             cores=dev_cores_aicore,
                             cores_utilization=dev_util_aicore,
                             memory=dev_mem,
                             memory_used=dev_mem_used,
-                            memory_utilization=(
-                                (dev_mem_used / dev_mem) * 100 if dev_mem > 0 else 0
-                            ),
+                            memory_utilization=get_utilization(dev_mem_used, dev_mem),
                             temperature=dev_temp,
-                            power_used=dev_power_used / 10,  # Convert from 0.1W to W
+                            power_used=dev_power_used,
                             appendix=dev_appendix,
                         ),
                     )
@@ -222,7 +220,11 @@ def _get_device_memory_info(dev_card_id, dev_device_id) -> tuple[int, int]:
             dev_mem = dev_memory_info.memory_size
             dev_mem_used = dev_memory_info.utiliza
     except pydcmi.DCMIError as e:
-        if e.value == pydcmi.DCMI_ERROR_FUNCTION_NOT_FOUND:
+        if e.value in [
+            pydcmi.DCMI_ERROR_FUNCTION_NOT_FOUND,
+            pydcmi.DCMI_ERROR_NOT_SUPPORT,
+            pydcmi.DCMI_ERROR_NOT_SUPPORT_IN_CONTAINER,
+        ]:
             dev_memory_info = pydcmi.dcmi_get_device_memory_info_v3(
                 dev_card_id,
                 dev_device_id,
@@ -260,8 +262,8 @@ def _get_device_roce_network_info(
             pydcmi.DCMI_PORT_TYPE_ROCE_PORT,
         )
     except pydcmi.DCMIError:
-        if logger.isEnabledFor(logging.DEBUG):
-            logger.exception("Failed to get device roce network info")
+        if slogger.isEnabledFor(logging.DEBUG):
+            slogger.exception("Failed to get device roce network info")
 
     return ip, mask, gateway
 
@@ -287,9 +289,113 @@ def _get_device_virtual_info(
             c_vdev_query_stru,
         )
     except pydcmi.DCMIError:
-        if logger.isEnabledFor(logging.DEBUG):
-            logger.exception("Failed to get device virtual info")
+        if slogger.isEnabledFor(logging.DEBUG):
+            slogger.exception("Failed to get device virtual info")
     else:
         return c_vdev_query_stru
 
     return None
+
+
+# Borrowed from https://gitcode.com/Ascend/pytorch/blob/master/torch_npu/csrc/core/npu/NpuVariables.cpp#L13-L40 and
+# https://gitcode.com/Ascend/pytorch/blob/master/torch_npu/csrc/core/npu/NpuVariables.h#L5-L34.
+_soc_name_version_mapping: dict[str, int] = {
+    "Ascend910PremiumA": 100,
+    "Ascend910ProA": 101,
+    "Ascend910A": 102,
+    "Ascend910ProB": 103,
+    "Ascend910B": 104,
+    "Ascend310P1": 200,
+    "Ascend310P2": 201,
+    "Ascend310P3": 202,
+    "Ascend310P4": 203,
+    "Ascend310P5": 204,
+    "Ascend310P7": 205,
+    "Ascend910B1": 220,
+    "Ascend910B2": 221,
+    "Ascend910B2C": 222,
+    "Ascend910B3": 223,
+    "Ascend910B4": 224,
+    "Ascend910B4-1": 225,
+    "Ascend310B1": 240,
+    "Ascend310B2": 241,
+    "Ascend310B3": 242,
+    "Ascend310B4": 243,
+    "Ascend910_9391": 250,
+    "Ascend910_9392": 251,
+    "Ascend910_9381": 252,
+    "Ascend910_9382": 253,
+    "Ascend910_9372": 254,
+    "Ascend910_9362": 255,
+}
+
+
+def _guess_soc_name_from_dev_name(dev_name: str) -> str | None:
+    """
+    Guess the SoC name from the device name.
+
+    Args:
+        dev_name:
+            The name of the device, e.g., "910A", "310P1", etc.
+
+    Returns:
+        The guessed SoC name, or None if not found.
+
+    """
+    soc_name = f"Ascend{dev_name}"
+    if soc_name in _soc_name_version_mapping:
+        return soc_name
+    return None
+
+
+def get_ascend_soc_version(name: str | None) -> int:
+    """
+    Get the Ascend SoC version based on the SoC name.
+
+    Args:
+        name:
+            The name of the SoC, e.g., "Ascend910A", "Ascend310P1", etc.
+
+    Returns:
+        The corresponding version number, or -1 if not found.
+
+    """
+    if not name:
+        return -1
+
+    version = _soc_name_version_mapping.get(name)
+    if version is None:
+        return -1
+
+    return version
+
+
+def get_ascend_cann_variant(name: str | None) -> str | None:
+    """
+    Get the CANN variant based on the SoC name.
+
+    Args:
+        name:
+            The name of the SoC, e.g., "Ascend910A", "Ascend310P1", etc.
+
+    Returns:
+        The corresponding cluster name, or None if not found.
+
+    """
+    if not name:
+        return None
+
+    version = get_ascend_soc_version(name)
+    if version <= 0:
+        return None
+    if version < 200:
+        return "910"
+    if version < 220:
+        return "310p"
+    if version < 240:
+        return "910b"
+    if version < 250:
+        return "310b"
+    if version < 260:
+        return "a3"  # 910c
+    return None
