@@ -5,6 +5,7 @@
 # https://rocm.docs.amd.com/projects/rocm_smi_lib/en/latest/doxygen/html/rocm__smi_8h.html.
 from __future__ import annotations
 
+import os
 import sys
 import threading
 from ctypes import *
@@ -18,19 +19,50 @@ ROCMSMI_ERROR_UNINITIALIZED = -99997
 rocmsmiLib = None
 libLoadLock = threading.Lock()
 
-if rocmsmiLib is None and Path("/opt/rocm/libexec/rocm_smi").exists():
-    libLoadLock.acquire()
+if rocmsmiLib is None:
+    # Example ROCM_SMI_LIB_PATH
+    # - /opt/dtk-24.04.3/rocm_smi/lib
+    # - /opt/rocm/rocm_smi/lib
+    rocmsmi_lib_path = os.getenv("ROCM_SMI_LIB_PATH")
+    if not rocmsmi_lib_path:
+        # Example ROCM_PATH/ROCM_HOME
+        # - /opt/dtk-24.04.3
+        # - /opt/rocm
+        rocm_path = Path(os.getenv("ROCM_HOME", os.getenv("ROCM_PATH") or "/opt/rocm"))
+        rocmsmi_lib_path = str(rocm_path / "lib")
+        if not Path(rocmsmi_lib_path).exists():
+            rocmsmi_lib_path = str(rocm_path / "rocm_smi" / "lib")
+    else:
+        rocm_path = Path(rocmsmi_lib_path).parent.parent
 
-    sys.path.append("/opt/rocm/libexec/rocm_smi/")
-    try:
-        # Refer to https://github.com/ROCm/rocm_smi_lib/blob/amd-staging_deprecated/python_smi_tools/rsmiBindings.py.
-        from rsmiBindings import *
+    rocmsmi_lib_loc = Path(rocmsmi_lib_path) / "librocm_smi64.so"
+    if rocmsmi_lib_loc.exists():
+        rocmsmi_bindings_paths = [
+            (rocm_path / "rocm_smi" / "bindings"),
+            (rocm_path / "libexec" / "rocm_smi"),
+        ]
+        rocmsmi_bindings_path = None
+        for p in rocmsmi_bindings_paths:
+            if p.exists():
+                rocmsmi_bindings_path = p
+                break
 
-        rocmsmiLib = initRsmiBindings()
-    except ImportError:
-        rocmsmiLib = None
-    finally:
-        libLoadLock.release()
+        if rocmsmi_bindings_path and rocmsmi_bindings_path.exists():
+            if str(rocmsmi_bindings_path) not in sys.path:
+                sys.path.append(str(rocmsmi_bindings_path))
+
+            libLoadLock.acquire()
+
+            try:
+                # Refer to https://github.com/ROCm/rocm_smi_lib/blob/amd-staging_deprecated/python_smi_tools/rsmiBindings.py.
+                from rsmiBindings import *
+
+                if not rocmsmiLib:
+                    rocmsmiLib = CDLL(rocmsmi_lib_loc)
+            except OSError:
+                pass
+            finally:
+                libLoadLock.release()
 
 
 class ROCMSMIError(Exception):
@@ -57,15 +89,6 @@ def _rocmsmiCheckReturn(ret):
 
 ## string/bytes conversion for ease of use
 def convertStrBytes(func):
-    """
-    In python 3, strings are unicode instead of bytes, and need to be converted for ctypes
-    Args from caller: (1, 'string', <__main__.c_nvmlDevice_t at 0xFFFFFFFF>)
-    Args passed to function: (1, b'string', <__main__.c_nvmlDevice_t at 0xFFFFFFFF)>.
-    ----
-    Returned from function: b'returned string'
-    Returned to caller: 'returned string'
-    """
-
     @wraps(func)
     def wrapper(*args, **kwargs):
         # encoding a str returns bytes in python 2 and 3
@@ -184,13 +207,21 @@ def rsmi_dev_target_graphics_version_get(device=0):
     if not rocmsmiLib:
         raise ROCMSMIError(ROCMSMI_ERROR_UNINITIALIZED)
 
-    c_version = c_uint64()
-    ret = rocmsmiLib.rsmi_dev_target_graphics_version_get(device, byref(c_version))
-    _rocmsmiCheckReturn(ret)
-    version = str(c_version.value)
-    if len(version) == 4:
-        version = hex(c_version.value)[2:]
-    return "gfx" + version
+    try:
+        c_version = c_uint64()
+        ret = rocmsmiLib.rsmi_dev_target_graphics_version_get(device, byref(c_version))
+        _rocmsmiCheckReturn(ret)
+        version = str(c_version.value)
+        if len(version) == 4:
+            dev_name = rsmi_dev_name_get(device)
+            if "Instinct MI2" in dev_name:
+                hex_part = str(hex(int(version[2:]))).replace("0x", "")
+                version = version[:2] + hex_part
+        else:
+            version = str(c_version.value // 10 + c_version.value % 10)
+        return "gfx" + version
+    except AttributeError:
+        return None
 
 
 def rsmi_dev_temp_metric_get(device=0, sensor=None, metric=None):
@@ -222,12 +253,37 @@ def rsmi_dev_power_cap_get(device=0):
     return c_power_cap.value // 1000000
 
 
+def rsmi_dev_power_ave_get(device=0):
+    if not rocmsmiLib:
+        raise ROCMSMIError(ROCMSMI_ERROR_UNINITIALIZED)
+
+    c_device_chip = c_uint32(0)
+    c_power = c_uint64()
+    ret = rocmsmiLib.rsmi_dev_power_ave_get(device, c_device_chip, byref(c_power))
+    _rocmsmiCheckReturn(ret)
+    return c_power.value // 1000000
+
+
 def rsmi_dev_power_get(device=0):
     if not rocmsmiLib:
         raise ROCMSMIError(ROCMSMI_ERROR_UNINITIALIZED)
 
-    c_power = c_uint64()
-    c_power_type = rsmi_power_type_t()
-    ret = rocmsmiLib.rsmi_dev_power_get(device, byref(c_power), byref(c_power_type))
+    try:
+        c_power = c_uint64()
+        c_power_type = rsmi_power_type_t()
+        ret = rocmsmiLib.rsmi_dev_power_get(device, byref(c_power), byref(c_power_type))
+        _rocmsmiCheckReturn(ret)
+        return c_power.value // 1000000
+    except NameError:
+        # Fallback for older versions without rsmi_dev_power_get
+        return rsmi_dev_power_ave_get(device)
+
+
+def rsmi_dev_node_id_get(device=0):
+    if not rocmsmiLib:
+        raise ROCMSMIError(ROCMSMI_ERROR_UNINITIALIZED)
+
+    c_node_id = c_uint32()
+    ret = rocmsmiLib.rsmi_dev_node_id_get(device, byref(c_node_id))
     _rocmsmiCheckReturn(ret)
-    return c_power.value // 1000000
+    return c_node_id.value
